#if !defined LIB_BLACKHOLE
#define LIB_BLACKHOLE
    const float c = 1.0;
    const float G = 8.51313773e-28;
    const float Msun = 1.989e30;
    const float bhMass = 1.7e8 * Msun;
    const float radiusSchwarzschild = (2.0 * G * bhMass) / (c * c);
    const float radiusIsco = 3.0 * radiusSchwarzschild;
    const float radiusDiskInner = radiusIsco;
    const float radiusDiskOuter = 8.0 * radiusIsco;

    const vec3 blackHolePosition = vec3(0.0, 0.0, -7e12);

    bool GravitationalField(inout vec3 rayPosition, inout vec3 rayDirection, in float timeStep) {
        vec3  singularityVector    = -(rayPosition.xyz - blackHolePosition);
        float singularityDistance  = length(singularityVector);

        vec3 g = 2.0 * G * bhMass * singularityVector 
                / (singularityDistance * singularityDistance * singularityDistance);

        rayDirection += timeStep * length(rayDirection) * g / c;
        rayDirection = normalize(rayDirection.xyz);
        rayPosition += rayDirection.xyz * c * timeStep;

        return length(rayPosition - blackHolePosition) <= 2.0;
    }

    float DopplerEffect(in float velocity) {
        return sqrt((1.0 + velocity) / (1.0 - velocity));
    }

    //From Zombye
    void AccretionDisk(
        vec3 rayPosition, vec3 rayDirection,
        out float density, out float radiance,
        in float wavelength
    ) {
        float r = length(rayPosition.xyz);

        if (r < radiusDiskInner) { return; }
        if (r > radiusDiskOuter) { return; }

        float phi = atan(rayPosition.y, rayPosition.x);

        float R10 = r * (100.0 / 1e10); // 10^10 centimeter
        float RStar = radiusIsco;
        float R = r;

        float Omega = sqrt(G*bhMass) / pow(r, 3.0 / 2.0);
        float vel = Omega * r;

        const float alpha = 1e-18;
        const float M16 = 5e-36 * bhMass * 1e3 / 1e16; // 10^16 grams
        const float m1 = bhMass / Msun; // solar mass

        float f = pow(1.0 - sqrt(radiusDiskInner / R), 1.0 / 4.0);

        float logTc_cp  = log2(1.4e+4) + log2(alpha) * (-1.0 /  5.0) + log2(M16) * ( 3.0 / 10.0) + log2(m1) * (1.0 / 4.0);
        float logRho_cp = log2(3.1e-8) + log2(alpha) * (-7.0 / 10.0) + log2(M16) * (11.0 / 20.0) + log2(m1) * (5.0 / 8.0);
        float logTc  = logTc_cp  + log2(R10) * ( -3.0 / 4.0) + log2(f) * ( 6.0 / 5.0);
        float logRho = logRho_cp + log2(R10) * (-15.0 / 8.0) + log2(f) * (11.0 / 5.0);

        float H   = 1.7e+8 * pow(alpha,-1.0 / 10.0) * pow(M16, 3.0 / 20.0) * pow(m1,-3.0 / 8.0) * pow(R10,  9.0 / 8.0) * pow(f,  3.0 / 5.0);
        float Tc  = exp2(logTc ); // kelvin
        float Rho = exp2(logRho); // gram / meter^3

        float noise = 1.0;//texture(iChannel1, vec2(r / radiusIsco, phi / tau)).r - 0.4;

        density = noise * Rho * 1e-7 * exp(-rayPosition.z * rayPosition.z / (0.2 * 0.2 * radiusSchwarzschild * radiusSchwarzschild));

        vec3 diskMotionVec = cross(normalize(rayPosition), vec3(0,0,1));
        vel *= dot(diskMotionVec, rayDirection);

        float diskTemperature = Tc;
        float dopplerFactor = DopplerEffect(vel);
        float redShift = sqrt(1.0 - radiusSchwarzschild / r);
        float diskEmission = Plancks(diskTemperature, dopplerFactor * wavelength * redShift) * pow(redShift * dopplerFactor, 5.0);

        radiance = diskEmission;
        if(isinf(radiance)) {
            radiance = 1.0;
        }
        if(isnan(radiance)) {
            radiance = 0.0;
        }
    }

    bool IntersectAccretionDisk(inout vec3 rayPosition, inout vec3 rayDirection, out float radiance, out float hitDistance, in float terrainDistance, in float wavelength) {
        /*
            Perform a biased march to find the intersection.
            Despite path tracing usually being unbiased, if the intersection used for it is able to converge on an exact intersection
            with an infinite number of steps then it is still path tracing.

            TODO: Figure out an unbiased march
        */

        vec2 accretionDists = RSI(rayPosition - blackHolePosition, rayDirection, radiusDiskOuter);
        vec2 blackHoleDists = RSI(rayPosition - blackHolePosition, rayDirection, radiusSchwarzschild);

        bool blackHoleIntersected = blackHoleDists.y >= 0.0;
        bool accretionIntersected = accretionDists.y >= 0.0;

        float stepSize = (2.0 * radiusDiskOuter) / (c * float(400));

        rayPosition = rayPosition + rayDirection * max(accretionDists.x, 0.0);

        float timeStep = stepSize;

        vec3 previousRayPosition = rayPosition;
        bool intersection = false;

        for(int i = 0; i < 400; ++i) {
            bool belowEventHorizon = GravitationalField(rayPosition, rayDirection, timeStep);
            if(belowEventHorizon) {
                break;
            }
            if(!accretionIntersected) {
                break;
            }

            float density;
            float tmp = 1.0;
            AccretionDisk(
                rayPosition - blackHolePosition, rayDirection,
                density, radiance,
                wavelength
            );

            if(!intersection) {
                hitDistance += length(rayPosition - previousRayPosition);
            }

            float stepDistance = length(rayPosition - previousRayPosition);

            intersection = exp(-density * min(terrainDistance - hitDistance, stepDistance)) < RandNextF();

            previousRayPosition = rayPosition;
            if(intersection) {
                break;
            }
        }

        return intersection;
    }
#endif