/*
====================================================================================================

        License Terms for Academy Color Encoding System Components

        Academy Color Encoding System (ACES) software and tools are provided by the
        Academy under the following terms and conditions: A worldwide, royalty-free,
        non-exclusive right to copy, modify, create derivatives, and use, in source and
        binary forms, is hereby granted, subject to acceptance of this license.

        Copyright © 2015 Academy of Motion Picture Arts and Sciences (A.M.P.A.S.).
        Portions contributed by others as indicated. All rights reserved.

        Performance of any of the aforementioned acts indicates acceptance to be bound
        by the following terms and conditions:

        * Copies of source code, in whole or in part, must retain the above copyright
        notice, this list of conditions and the Disclaimer of Warranty.

        * Use in binary form must retain the above copyright notice, this list of
        conditions and the Disclaimer of Warranty in the documentation and/or other
        materials provided with the distribution.

        * Nothing in this license shall be deemed to grant any rights to trademarks,
        copyrights, patents, trade secrets or any other intellectual property of
        A.M.P.A.S. or any contributors, except as expressly stated herein.

        * Neither the name "A.M.P.A.S." nor the name of any other contributors to this
        software may be used to endorse or promote products derivative of or based on
        this software without express prior written permission of A.M.P.A.S. or the
        contributors, as appropriate.

        This license shall be construed pursuant to the laws of the State of
        California, and any disputes related thereto shall be subject to the
        jurisdiction of the courts therein.

        Disclaimer of Warranty: THIS SOFTWARE IS PROVIDED BY A.M.P.A.S. AND CONTRIBUTORS
        "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
        THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, AND
        NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL A.M.P.A.S., OR ANY
        CONTRIBUTORS OR DISTRIBUTORS, BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
        SPECIAL, EXEMPLARY, RESITUTIONARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
        LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
        PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
        LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
        OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
        ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

        WITHOUT LIMITING THE GENERALITY OF THE FOREGOING, THE ACADEMY SPECIFICALLY
        DISCLAIMS ANY REPRESENTATIONS OR WARRANTIES WHATSOEVER RELATED TO PATENT OR
        OTHER INTELLECTUAL PROPERTY RIGHTS IN THE ACADEMY COLOR ENCODING SYSTEM, OR
        APPLICATIONS THEREOF, HELD BY PARTIES OTHER THAN A.M.P.A.S.,WHETHER DISCLOSED OR
        UNDISCLOSED.

====================================================================================================
*/

#include "constants.glsl"

vec3 xyY_2_XYZ(vec3 xyY)
{
	vec3 XYZ;
	XYZ[0] = xyY[0] * xyY[2] / max( xyY[1], 1e-10);
	XYZ[1] = xyY[2];  
	XYZ[2] = (1.0 - xyY[0] - xyY[1]) * xyY[2] / max( xyY[1], 1e-10);

	return XYZ;
}

mat3 chromatic_adaption(vec2 src_xy, vec2 dst_xy)
{
	// Von Kries chromatic adaptation 

	// Bradford
	const mat3 coneResponse =
	mat3(
		 0.8951,  0.2664, -0.1614,
		-0.7502,  1.7135,  0.0367,
		 0.0389, -0.0685,  1.0296
    );
	const mat3 invConeResponse =
	mat3(
		 0.9869929, -0.1470543,  0.1599627,
		 0.4323053,  0.5183603,  0.0492912,
		-0.0085287,  0.0400428,  0.9684867
	);

	vec3 src_XYZ = xyY_2_XYZ(vec3(src_xy, 1));
	vec3 dst_XYZ = xyY_2_XYZ(vec3(dst_xy, 1));

	vec3 srcConeResp = (coneResponse * src_XYZ);
	vec3 dstConeResp = (coneResponse * dst_XYZ);

	mat3 vonKriesMat =
	mat3(
		dstConeResp[0] / srcConeResp[0], 0.0, 0.0,
		0.0, dstConeResp[1] / srcConeResp[1], 0.0,
		0.0, 0.0, dstConeResp[2] / srcConeResp[2]
    );

	return (invConeResponse * (vonKriesMat * coneResponse));
}

vec2 plankianLocusChromaticity(float temp)
{
	float u = (0.860117757f + 1.54118254e-4f * temp + 1.28641212e-7f * temp*temp) / (1.0f + 8.42420235e-4f * temp + 7.08145163e-7f * temp*temp);
	float v = (0.317398726f + 4.22806245e-5f * temp + 4.20481691e-8f * temp*temp) / (1.0f - 2.89741816e-5f * temp + 1.61456053e-7f * temp*temp);

	float x = 3*u / ( 2*u - 8*v + 4 );
	float y = 2*v / ( 2*u - 8*v + 4 );

	return vec2(x,y);
}

// Accurate for 4000K < temp < 25000K
// in: correlated color temperature
// out: CIE 1931 chromaticity
vec2 dIlluminantChromaticity(float temp)
{
	// Correct for revision of Plank's law
	// This makes 6500 == D65
	temp *= 1.4388 / 1.438;

	float x =	temp <= 7000 ?
				0.244063 + (0.09911e3 + (2.9678e6 - 4.6070e9 / temp) / temp) / temp :
				0.237040 + (0.24748e3 + (1.9018e6 - 2.0064e9 / temp) / temp) / temp;
	
	float y = -3 * x*x + 2.87 * x - 0.275;

	return vec2(x,y);
}

// Find closest color temperature to chromaticity
// [McCamy 1992, "Correlated color temperature as an explicit function of chromaticity coordinates"]
float correlatedColorTemperature(float x, float y)
{
	float n = (x - 0.3320) / (0.1858 - y);
	return -449 * n*n*n + 3525 * n*n - 6823.3 * n + 5520.33;
}

vec2 plankianIsothermal(float temp, float Tint)
{
	float u = (0.860117757f + 1.54118254e-4f * temp + 1.28641212e-7f * temp*temp) / (1.0f + 8.42420235e-4f * temp + 7.08145163e-7f * temp*temp);
	float v = (0.317398726f + 4.22806245e-5f * temp + 4.20481691e-8f * temp*temp) / (1.0f - 2.89741816e-5f * temp + 1.61456053e-7f * temp*temp);

	float ud = (-1.13758118e9f - 1.91615621e6f * temp - 1.53177f * temp*temp) / square(1.41213984e6f + 1189.62f * temp + temp*temp);
	float vd = (1.97471536e9f - 705674.0f * temp - 308.607f * temp*temp) / square(6.19363586e6f - 179.456f * temp + temp*temp);

	vec2 uvd = normalize(vec2(u, v));

	// Correlated color temperature is meaningful within +/- 0.05
	u += -uvd.y * Tint * 0.05;
	v +=  uvd.x * Tint * 0.05;
	
	float x = 3*u / ( 2*u - 8*v + 4 );
	float y = 2*v / ( 2*u - 8*v + 4 );

	return vec2(x,y);
}

float whiteTemp = WHITE_TEMP;
float whiteTint = 0.0;

vec3 WhiteBalance(vec3 linearColor)
{
	vec2 srcWhiteDaylight = dIlluminantChromaticity(whiteTemp);
	vec2 srcWhitePlankian = plankianLocusChromaticity(whiteTemp);

	vec2 srcWhite = whiteTemp < 4000 ? srcWhitePlankian : srcWhiteDaylight;
	vec2 d65_white = vec2(0.31270,  0.32900);

	{
		// Offset along isotherm
		vec2 isothermal = plankianIsothermal(whiteTemp, whiteTint) - srcWhitePlankian;
		srcWhite += isothermal;
	}

	mat3 whiteBalanceMat = chromatic_adaption(srcWhite, d65_white);
	whiteBalanceMat = (sRGB_2_XYZ_MAT * whiteBalanceMat) * XYZ_2_sRGB_MAT;

	return linearColor * whiteBalanceMat;
}

float rgb_2_saturation( vec3 rgb )
{
	float minrgb = min( min(rgb.r, rgb.g ), rgb.b );
	float maxrgb = max( max(rgb.r, rgb.g ), rgb.b );
	return ( max( maxrgb, 1e-10 ) - max( minrgb, 1e-10 ) ) / max( maxrgb, 1e-2 );
}

float glow_fwd( float yc_in, float glow_gain_in, float glow_mid)
{
   float glow_gain_out;

   if (yc_in <= 2./3. * glow_mid) {
	 glow_gain_out = glow_gain_in;
   } else if ( yc_in >= 2 * glow_mid) {
	 glow_gain_out = 0;
   } else {
	 glow_gain_out = glow_gain_in * (glow_mid / yc_in - 0.5);
   }

   return glow_gain_out;
}

float glow_inv( float ycOut, float glow_gain_in, float glow_mid)
{
	float glow_gain_out;

	if (ycOut <= ((1 + glow_gain_in) * 2./3. * glow_mid)) {
		glow_gain_out = -glow_gain_in / (1 + glow_gain_in);
	} else if ( ycOut >= (2. * glow_mid)) {
		glow_gain_out = 0.;
	} else {
		glow_gain_out = glow_gain_in * (glow_mid / ycOut - 1./2.) / (glow_gain_in / 2. - 1.);
	}

	return glow_gain_out;
}


float sigmoid_shaper( float x)
{
	float t = max( 1 - abs( 0.5 * x ), 0 );
	float y = 1 + sign(x) * (1 - t*t);
	return 0.5 * y;
}

float rgb_hue( vec3 rgb ) 
{
	float hue;
	if (rgb[0] == rgb[1] && rgb[1] == rgb[2])
	{
		hue = 0;
	}
	else
	{
		hue = (180.0 / pi) * atan(2.0 * rgb[0] - rgb[1] - rgb[2], sqrt(3.0) * (rgb[1] - rgb[2]));
	}

	if (hue < 0.)
		hue = hue + 360;

	return clamp( hue, 0, 360 );
}

float rgb_yc( vec3 rgb, float ycRadiusWeight)
{
	float r = rgb[0]; 
	float g = rgb[1]; 
	float b = rgb[2];
  
	float chroma = sqrt(b*(b-g)+g*(g-r)+r*(r-b));

	return ( b + g + r + ycRadiusWeight * chroma) / 3.;
}

float rgb_yc( vec3 rgb)
{
return rgb_yc(rgb, 1.75);
}

float sigmoid_sharper(float x) {
    float t     = max(1.0 - abs(0.5 * x), 0);
    float y     = 1.0 + sign(x) * (1.0 - square(t));

    return 0.5 * y;
}

float cubic_basis_shaper(float x, float w) {

    const vec4 M[4] = vec4[4](
        vec4( -1.0 / 6.0,  3.0 / 6.0, -3.0 / 6.0,  1.0 / 6.0),
        vec4(  3.0 / 6.0, -6.0 / 6.0,  3.0 / 6.0,  0.0 / 6.0), 
        vec4( -3.0 / 6.0,  0.0 / 6.0,  3.0 / 6.0,  0.0 / 6.0), 
        vec4(  1.0 / 6.0,  4.0 / 6.0,  1.0 / 6.0,  0.0 / 6.0)
    );

    float knot[5] = float[5](
        -0.5 * w,
        -0.25 * w,
        0.0, 
        0.25 * w,
        0.5 * w
    );

    float y     = 0.0;

    if ((x > knot[0]) && (x < knot[4])) {
        float knot_coord    = (x - knot[0]) * 4.0 * rcp(w);

        int j   = int(knot_coord);
        float t = knot_coord - j;

        float monomials[4]  = float[4](
            cube(t),
            square(t),
            t,
            1.0
        );

        /* this section can be simplified */
        if (j == 3) {
            y   = monomials[0] * M[0].x + monomials[1] * M[1].x
                + monomials[2] * M[2].x + monomials[3] * M[3].x;
        } else if (j == 2) {
            y   = monomials[0] * M[0].y + monomials[1] * M[1].y
                + monomials[2] * M[2].y + monomials[3] * M[3].y;
        } else if (j == 1) {
            y   = monomials[0] * M[0].z + monomials[1] * M[1].z
                + monomials[2] * M[2].z + monomials[3] * M[3].z;
        } else if (j == 0) {
            y   = monomials[0] * M[0].w + monomials[1] * M[1].w
                + monomials[2] * M[2].w + monomials[3] * M[3].w;
        } else {
            y   = 0.0;
        }
    }

    return y * 1.5;
}

const mat3 M = mat3(
    0.5, -1.0, 0.5,
    -1.0, 1.0, 0.5, 
    0.5, 0.0, 0.0
);

struct segmented_spline_param_c5 {
    float coeff_low[6];
    float coeff_high[6];
    vec2 min_point;
    vec2 mid_point;
    vec2 max_point;
    float slope_low;
    float slope_high;
};

float segmented_spline_c5_fwd(float x) {
    /* rrt params */
    const segmented_spline_param_c5 c = segmented_spline_param_c5(
        float[6](-4.0, -4.0, -3.1573765773, -0.4852499958, 1.8477324706, 1.8477324706), 
        float[6](-0.7185482425, 2.0810307172, 3.6681241237, 4.0, 4.0, 4.0),
        vec2(0.18 * exp2(-18.0), 0.0001),
        vec2(0.18, 4.8),
        vec2(0.18 * exp2(18.0), 10000.0),
        0.0,
        0.0
    );

    const int N_KNOTS_LOW   = 4;
    const int N_KNOTS_HIGH  = 4;

    float x_check = x <= 0.0 ? exp2(-14.0) : x;

    float logx  = log10(x_check);
    float logy;

    if (logx <= log10(c.min_point.x)) {
        logy    = logx * c.slope_low + (log10(c.min_point.y) - c.slope_low * log10(c.min_point.x));
    } else if ((logx > log10(c.min_point.x)) && (logx < log10(c.mid_point.x))) {
        float knot_coord = (N_KNOTS_LOW - 1) * (logx - log10(c.min_point.x)) * rcp(log10(c.mid_point.x) - log10(c.min_point.x));
        int j   = int(knot_coord);
        float t = knot_coord - j;

        vec3 cf = vec3(c.coeff_low[j], c.coeff_low[j+1], c.coeff_low[j+2]);

        vec3 monomials = vec3(square(t), t, 1.0);

        logy    = dot(monomials, M * cf);
    } else if ((logx >= log10(c.mid_point.x)) && (logx < log10(c.max_point.x))) {
        float knot_coord = (N_KNOTS_HIGH - 1) * (logx - log10(c.mid_point.x)) * rcp(log10(c.max_point.x) - log10(c.mid_point.x));
        int j   = int(knot_coord);
        float t = knot_coord - j;

        vec3 cf = vec3(c.coeff_high[j], c.coeff_high[j+1], c.coeff_high[j+2]);

        vec3 monomials = vec3(square(t), t, 1.0);

        logy    = dot(monomials, M * cf);
    } else {
        logy    = logx * c.slope_high + (log10(c.max_point.y) - c.slope_high * log10(c.max_point.x));
    }
    return pow(10.0, logy);
}

struct segmented_spline_param_c9 {
    float coeff_low[10];
    float coeff_high[10];
    vec2 min_point;
    vec2 mid_point;
    vec2 max_point;
    float slope_low;
    float slope_high;
};

float segmented_spline_c9_fwd(float x, const segmented_spline_param_c9 c) {
    const int N_KNOTS_LOW   = 8;
    const int N_KNOTS_HIGH  = 8;

    float x_check = x <= 0.0 ? 1e-4 : x;

    float logx  = log10(x_check);
    float logy;

    if (logx <= log10(c.min_point.x)) {
        logy    = logx * c.slope_low + (log10(c.min_point.y) - c.slope_low * log10(c.min_point.x));
    } else if ((logx > log10(c.min_point.x)) && (logx < log10(c.mid_point.x))) {
        float knot_coord = (N_KNOTS_LOW - 1) * (logx - log10(c.min_point.x)) * rcp(log10(c.mid_point.x) - log10(c.min_point.x));
        int j   = int(knot_coord);
        float t = knot_coord - j;

        vec3 cf = vec3(c.coeff_low[j], c.coeff_low[j+1], c.coeff_low[j+2]);

        vec3 monomials = vec3(square(t), t, 1.0);

        logy    = dot(monomials, M * cf);
    } else if ((logx >= log10(c.mid_point.x)) && (logx < log10(c.max_point.x))) {
        float knot_coord = (N_KNOTS_HIGH - 1) * (logx - log10(c.mid_point.x)) * rcp(log10(c.max_point.x) - log10(c.mid_point.x));
        int j   = int(knot_coord);
        float t = knot_coord - j;

        vec3 cf = vec3(c.coeff_high[j], c.coeff_high[j+1], c.coeff_high[j+2]);

        vec3 monomials = vec3(square(t), t, 1.0);

        logy    = dot(monomials, M * cf);
    } else {
        logy    = logx * c.slope_high + (log10(c.max_point.y) - c.slope_high * log10(c.max_point.x));
    }
    return pow(10.0, logy);
}

float center_hue(float hue, float center) {
    float hue_centered = hue - center;

    if (hue_centered < -180.0) hue_centered += 360.0;
    else if (hue_centered > 180.0) hue_centered -= 360.0;

    return hue_centered;
}

float y_linVC(float y, float max, float min) {
    return (y - min) / (max - min);
}

vec3 XYZ_xyY(vec3 xyz) {
    vec3 xyY;
    float divisor   = (xyz.r + xyz.g + xyz.b);

    if (divisor == 0.0) divisor = 1e-10;
    xyY.r   = xyz.r / divisor;
    xyY.g   = xyz.g / divisor;
    xyY.b   = xyz.g;

    return xyY;
}
vec3 xyY_XYZ(vec3 xyY) {
    vec3 xyz;

    xyz.r   = xyY.x * xyY.z / max(xyY.g, 1e-10);
    xyz.g   = xyY.z;
    xyz.b   = (1.0 - xyY.x - xyY.y) * xyY.z / max(xyY.y, 1e-10);

    return xyz;
}

const float dim_surround_gamma = 0.9811;     //default 0.9811

vec3 dark_to_dim_surround(vec3 linear_cv) {
    vec3 xyz    = linear_cv * AP1_2_XYZ_MAT;
    vec3 xyY    = XYZ_xyY(xyz);
        xyY.z   = clamp(xyY.z, 0.0, 65535.0);
        xyY.z   = pow(xyY.z, dim_surround_gamma);
        xyz     = xyY_XYZ(xyY);

    return xyz * XYZ_2_AP1_MAT;
}

/* takes input as academy color and returns oces */
vec3 academy_rrt(vec3 aces) {
    /* glow module */
    const float rrt_glow_gain   = 0.21;     //default 0.05
    const float rrt_glow_mid    = 0.08;     //default 0.08

    float saturation    = rgb_2_saturation(aces);
    float yc_in         = rgb_yc(aces);
    float s             = sigmoid_sharper((saturation - 0.4) * rcp(0.2));
    float added_glow    = 1.0 + glow_fwd(yc_in, rrt_glow_gain * s, rrt_glow_mid);

    aces   *= added_glow;

    /* red modifier module */
    const float rrt_red_scale = 0.82;       //default 0.82
    const float rrt_red_privot = 0.03;      //default 0.03
    const float rrt_red_hue = 0.0;          //default 0.0
    const float rrt_red_width = 135.0;      //default 135.0

    float hue           = rgb_hue(aces);
    float centered_hue  = center_hue(hue, rrt_red_hue);
    float hue_weight    = cubic_basis_shaper(centered_hue, rrt_red_width);

        aces.r += hue_weight * saturation * (rrt_red_privot - aces.r) * (1.0 - rrt_red_scale);

    /* aces to rgb */
        aces    = clamp(aces, 0.0, 65535.0);

        vec3 rgb_pre    = aces * AP0_AP1;
            rgb_pre     = clamp(rgb_pre, 0.0, 65535.0);

    /* global desaturation */
    const float rrt_sat_factor = 0.96;  //default 0.96
        rgb_pre         = mix(vec3(dot(rgb_pre, AP1_RGB2Y)), rgb_pre, rrt_sat_factor);

    /* apply tonescale in rgb space */
    vec3 rgb_post;
        rgb_post.r      = segmented_spline_c5_fwd(rgb_pre.x);
        rgb_post.g      = segmented_spline_c5_fwd(rgb_pre.y);
        rgb_post.b      = segmented_spline_c5_fwd(rgb_pre.b);

    /* rgb space to oces */
    return max(rgb_post, 0.0) * AP1_AP0;
}

vec3 odt_srgb_d65(vec3 oces) {
    vec3 rgb_pre = oces * AP0_AP1;

    segmented_spline_param_c9 odt_48nit = segmented_spline_param_c9(
        float[10](-1.6989700043, -1.6989700043, -1.4779000000, -1.2291000000, -0.8648000000, -0.4480000000, 0.0051800000, 0.4511080334, 0.9113744414, 0.9113744414),
        float[10](0.5154386965, 0.8470437783, 1.1358000000, 1.3802000000, 1.5197000000, 1.5985000000, 1.6467000000, 1.6746091357, 1.6878733390, 1.6878733390),
        vec2(segmented_spline_c5_fwd(0.18 * exp2(-6.5)), 0.02),
        vec2(segmented_spline_c5_fwd(0.18), 4.8),
        vec2(segmented_spline_c5_fwd(0.18 * exp2(6.5)), 48.0),
        0.0,
        0.04
    );

    /* apply tonescale in rgb space */
    vec3 rgb_post;
        rgb_post.r      = segmented_spline_c9_fwd(rgb_pre.x, odt_48nit);
        rgb_post.g      = segmented_spline_c9_fwd(rgb_pre.y, odt_48nit);
        rgb_post.b      = segmented_spline_c9_fwd(rgb_pre.b, odt_48nit);

    /* black and white points for cinema system */
    const float cinema_white    = 48.0;
    const float cinema_black    = cinema_white / 2400.0;     //white / 2400.0, default 0.02

    /* scale luminance to linear */
    vec3 linear_cv;
        linear_cv.r     = y_linVC(rgb_post.r, cinema_white, cinema_black);
        linear_cv.g     = y_linVC(rgb_post.g, cinema_white, cinema_black);
        linear_cv.b     = y_linVC(rgb_post.b, cinema_white, cinema_black);

        linear_cv       = dark_to_dim_surround(linear_cv);

    /* global desaturation */
    const float odt_sat_factor = 0.96;  //default 0.96
        linear_cv       = mix(vec3(dot(linear_cv, AP1_RGB2Y)), linear_cv, odt_sat_factor);

    /* convert to display primary encoding */
    vec3 xyz    = linear_cv * AP1_2_XYZ_MAT;
        xyz     = xyz * D60_2_D65_CAT;

        linear_cv   = xyz * XYZ_2_sRGB_MAT;

    return saturate(linear_cv);
}

vec3 academy_color_encoding_sytem_rrt(vec3 linear) {

    vec3 aces = linear * sRGB_2_AP0;
         aces = WhiteBalance(aces);

    vec3 oces = academy_rrt(aces);
         oces = odt_srgb_d65(oces);
         oces = saturate(oces);

    return max(oces, 0.0);
}
