#if !defined BSDF_CLEARCOAT
#define BSDF_CLEARCOAT
    /*
        Calculate the conductive multiple scattering BSDF.

        This BSDF model handles a two interface BSDF. The first interface is always dielectric, and the final interface can be either opaque dielectric or conductor.
        It's polished materials and wet materials.
    */
    const int maxPathSegments = 50;
    const int mediaCount = 3;
    const int interfaceCount = mediaCount - 1;

    vec3 SampleInterface(
            in vec3 incomingDirection, 
            inout float estimate,
            inout float throughput, 
            in float wavelength,
            in int cInterface,
            in Material bounceMaterial,
            in Material clearCoatMaterial,
            in VoxelData data,
            out bool outside
        ) {
        if(cInterface == interfaceCount - 1) {
            //Bottom interface

            float conductorProbability = float(bounceMaterial.isMetal);
            bool conductor = conductorProbability > RandNextF();
            //Use the normal mapped BSDF. Should probably switch to the microfacet based normal mapping version.
            if(conductor) {
                return SampleConductor_MBN(incomingDirection, throughput, wavelength, data, bounceMaterial);
            } else {
                return SampleOpaqueDielectric_MBN(incomingDirection, estimate, throughput, wavelength, data, bounceMaterial);
            }
        } else {
            return SampleDielectric(incomingDirection, throughput, outside, wavelength, clearCoatMaterial);
        }
    }

    vec3 SampleClearCoat(in vec3 incomingDirection, inout float estimate, inout float throughput, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        int currentInterface = (incomingDirection.z < 0.0) ? 0 : interfaceCount - 1;

        float wavelengthMix = (wavelength - 390.0) / 441.0;

        float albedoSpectra;
        RGBToSpectrum(albedoSpectra, wavelength, 0.3, 0.6, 0.7, 0);

        Material clearCoatMaterial = bounceMaterial;
        clearCoatMaterial.albedo = albedoSpectra;
        clearCoatMaterial.radius = 0.000002;
        clearCoatMaterial.roughness = 0.0;
        clearCoatMaterial.n2 = complexFloat(mix(1.3, 2.4, wavelengthMix), 0.0);

        bounceMaterial.n1 = clearCoatMaterial.n2;

        vec3 newDirection = incomingDirection;

        Volume layer = Volume(0.0, 0.0, 0.0, 0.0, 0.0);

        int iteration = 0;
        do {
            iteration++;
            if(iteration > maxPathSegments) break;

            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            bool outside = true;
            newDirection = SampleInterface(
                newDirection, 
                estimate,
                throughput, 
                wavelength,
                currentInterface,
                bounceMaterial,
                clearCoatMaterial,
                data,
                outside
            );

            if(!outside) {
                layer = DefineSubsurface(clearCoatMaterial.albedo, clearCoatMaterial.radius, wavelength, 0);
            } else {
                layer = Volume(0.0, 0.0, 0.0, 0.0, 0.0);
            }

            if (newDirection.z > 0.0) {
                if(currentInterface != 0) {
                    float transmittance = pow(saturate(albedoSpectra), 1.0 / abs(newDirection.z));
                    throughput *= transmittance;

                    //if(RandNextF() < layer.albedo) {
                    //    newDirection = Rotate(GenerateUnitVector(RandNext2F()), vec3(0.0, 0.0, 1.0), newDirection);
                    //} else {
                    //    break;
                    //}
                }

                --currentInterface;
            } else if (newDirection.z < 0.0) {
                if(currentInterface != interfaceCount - 1) {
                    float transmittance = pow(saturate(albedoSpectra), 1.0 / abs(newDirection.z));
                    throughput *= transmittance;

                    //if(RandNextF() < layer.albedo) {
                    //    newDirection = Rotate(GenerateUnitVector(RandNext2F()), vec3(0.0, 0.0, 1.0), newDirection);
                    //} else {
                    //    break;
                    //}
                }

                ++currentInterface;
            } else {
                break;
            }
        } while (currentInterface >= 0 && currentInterface < interfaceCount);
        
        return newDirection;
    }
#endif