#if !defined BSDF_CONDUCTOR
#define BSDF_CONDUCTOR
    /*
        Calculate the conductive multiple scattering BSDF.

        This BSDF model handles purely specular reflection.
        It's metals.
        This doesn't actually need phase functions or facet BSDFs, it has them for the sake of consistency.
    */

    vec3 SampleConductorFacetBSDF(in vec3 direction, in vec3 normal, inout float throughput, in float wavelength, in Material bounceMaterial) {
        float fresnel = FresnelNonPolarized_R(dot(normal, direction), bounceMaterial.n1, bounceMaterial.n2);

        if(bounceMaterial.thinFilm) {
            fresnel = FresnelThinFilmInterferenceReflected(dot(normal, direction), bounceMaterial.thickness, wavelength, bounceMaterial.n1, bounceMaterial.filmIOR, bounceMaterial.n2);
        }

        throughput *= fresnel;

        if(isnan(throughput)) throughput = 1.0;

        return reflect(-direction, normal);
    }

    vec3 SampleConductor(in vec3 direction, inout float throughput, in float wavelength, in Material bounceMaterial) {
        int scatteringOrder = 0;
        float hr = 1.0 + InvC1(0.999);
        while(scatteringOrder <= MAX_BSDF_SCATTERING_EVENTS) {
            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            float U = RandNextF();

            vec3 wm = SampleD_wi(-direction, RandNext2F(), vec2(bounceMaterial.roughness));

            hr = SampleHeight(direction, vec2(bounceMaterial.roughness), hr, U);

            if(hr == 1e35) {
                break;
            }

            direction = SampleConductorFacetBSDF(-direction, wm, throughput, wavelength, bounceMaterial);

            if(hr != hr || direction.z != direction.z) {
                return vec3(0, 0, 1);
            }
            scatteringOrder++;
        }

        if(isnan(throughput)) {
            throughput = 0.0;
        }

        if(scatteringOrder > MAX_BSDF_SCATTERING_EVENTS) {
            throughput = 0.0;
        }

        return direction;
    }

    float EstimateConductorPhaseFunction(in vec3 incomingDirection, in vec3 outgoingDirection, in vec2 alpha, in float wavelength, in Material bounceMaterial) {
        vec3 halfway = normalize(incomingDirection + outgoingDirection);

        float vDotH = dot(incomingDirection, halfway);

        float specular  = FresnelNonPolarized_R(vDotH, bounceMaterial.n1, bounceMaterial.n2);

        specular *= D_wi(incomingDirection, halfway, alpha);
        specular /= 4.0 * vDotH;

        return specular;
    }

    float EstimateConductor(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounce, in float wavelength, in Material bounceMaterial) {
        float hr = 1.0 + InvC1(0.999);

        vec2 alpha = vec2(bounceMaterial.roughness);

        #ifdef CELESTIAL_PATH_ROUGHNESS_BOUNCE_INCREASE
            if(bounce > 1) {
                alpha = 0.01;
                alpha *= 15.0;
                alpha = saturate(alpha);
            }
        #endif

        bool d1_outside = incomingDirection.z < 0.0;
        bool d2_outside = outgoingDirection.z > 0.0;

        if (d1_outside != d2_outside) { return 0.0; }

        int scatteringOrder = 0;
        float throughput = 1.0;
        float estimate = 0.0;
        while(scatteringOrder <= MAX_BSDF_SCATTERING_EVENTS) {
            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            float U = RandNextF();
            hr = SampleHeight(incomingDirection, alpha, hr, U);

            if(hr == 1e35) {
                break;
            }

            scatteringOrder++;

            float phase = EstimateConductorPhaseFunction(-incomingDirection, outgoingDirection, alpha, wavelength, bounceMaterial);
            if(isnan(phase)) {
                phase = 1.0;
            }
            estimate += throughput * phase * G_1(outgoingDirection, alpha, hr);

            vec3 wm = SampleD_wi(-incomingDirection, RandNext2F(), alpha);

            if(isnan(throughput)) {
                throughput = 0.0;
            }

            float tmp;
            incomingDirection = SampleConductorFacetBSDF(-incomingDirection, wm, throughput, wavelength, bounceMaterial);

            if(hr != hr || incomingDirection.z != incomingDirection.z) {
                return 0.0;
            }
        }
        if(isnan(estimate)) {
            estimate = 0.0;
        }

        return estimate / outgoingDirection.z;
    }
#endif