#if !defined BSDF_DIELECTRIC
#define BSDF_DIELECTRIC
    /*
        Calculate the dielectric multiple scattering BSDF.

        This BSDF model handles a transparent dielectric material. All light that enters has a chance to exit on the opposite side.
        Glass and other refractive materials fall under this BSDF.
    */
    vec3 SampleDielectricPhaseFunction(in vec3 direction, inout bool outside, inout float throughput, in float wavelength, in Material bounceMaterial) {
        vec2 alpha = vec2(bounceMaterial.roughness);

        vec3 wm = outside ? SampleD_wi(direction, RandNext2F(), alpha) : -SampleD_wi(-direction, RandNext2F(), alpha);

        complexFloat n1 = bounceMaterial.n1;
        complexFloat n2 = bounceMaterial.n2;

        if(!outside) {
            n1 = bounceMaterial.n2;
            n2 = bounceMaterial.n1;
        }

        float fresnel_R = FresnelNonPolarized_R(dot(wm, direction), n1, n2);
        float fresnel_T = FresnelNonPolarized_T(dot(wm, direction), n1, n2);

        float totalLum = fresnel_R;
        float specBounceProbability = fresnel_R;

        bool specularBounce = specBounceProbability > RandNextF();

        if(specularBounce) {
            throughput /= specBounceProbability;

            throughput *= fresnel_R;
            throughput  = abs(throughput);

            return reflect(-direction, wm);
        } else {
            outside = !outside;

            throughput /= 1.0 - specBounceProbability;

            throughput *= fresnel_T;
            throughput  = abs(throughput);

            return refractIT(-direction, wm, n1.r, n2.r);
        }
    }

    vec3 SampleDielectric(in vec3 direction, inout float throughput, inout bool outside, in float wavelength, in Material bounceMaterial) {
        int scatteringOrder = 0;
        float hr = 1.0 + InvC1(0.999);
        
        outside = direction.z < 0.0;

        if (!outside) { hr = -hr; }

        while(scatteringOrder <= MAX_BSDF_SCATTERING_EVENTS) {
            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            float U = RandNextF();
            hr = outside ? SampleHeight(direction, vec2(bounceMaterial.roughness), hr, U) : -SampleHeight(-direction, vec2(bounceMaterial.roughness), -hr, U);

            if(hr == 1e35 || hr == -1e35) {
                break;
            }

            scatteringOrder++;

            direction = SampleDielectricPhaseFunction(-direction, outside, throughput, wavelength, bounceMaterial);

            if(hr != hr || direction.z != direction.z) {
                return vec3(0, 0, 1);
            }
        }

        if(scatteringOrder > MAX_BSDF_SCATTERING_EVENTS) {
            throughput = 0.0;
        }

        return direction;
    }

    float EstimateDielectricPhase(in vec3 incomingDirection, in vec3 outgoingDirection, in vec2 alpha, in bool incomingOutside, in bool outgoingOutside, in float wavelength, in Material bounceMaterial) {
        complexFloat n1 = bounceMaterial.n1;
        complexFloat n2 = bounceMaterial.n2;

        if(!incomingOutside) {
            n1 = bounceMaterial.n2;
            n2 = bounceMaterial.n1;
        }

        float eta = n1.r / n2.r;

        if(incomingOutside == outgoingOutside) {
            vec3 halfway = normalize(incomingDirection + outgoingDirection);

            float vDotH = incomingOutside ? dot(incomingDirection, halfway) : dot(-incomingDirection, -halfway);

            float numerator = incomingOutside ?
            D_wi( incomingDirection,  halfway, alpha) * FresnelNonPolarized_R(dot( incomingDirection,  halfway), bounceMaterial.n1, bounceMaterial.n2) :
            D_wi(-incomingDirection, -halfway, alpha) * FresnelNonPolarized_R(dot(-incomingDirection, -halfway), bounceMaterial.n1, bounceMaterial.n2);
            float denominator = 4.0 * vDotH;
            float specular = numerator / denominator;

            return specular;
        } else {
            vec3 tmp = incomingDirection;
            incomingDirection = outgoingDirection;
            outgoingDirection = tmp;
            incomingDirection = -incomingDirection;
            outgoingDirection = -outgoingDirection;

            vec3 halfway = -normalize(incomingDirection + outgoingDirection * eta);
            halfway *= incomingOutside ? sign(halfway.z) : -sign(halfway.z);

            float refraction = 0.0;
            #ifndef FORCE_APPROXIMATE_ESTIMATE_DIELECTRIC
                if(incomingOutside) {
                    refraction = eta * eta * FresnelNonPolarized_T(dot(incomingDirection, halfway), bounceMaterial.n1, bounceMaterial.n2) *
                                D_wi(incomingDirection, halfway, alpha) * max(0.0, -dot(outgoingDirection, halfway)) *
                                1.0 / square(dot(incomingDirection, halfway) + eta * dot(outgoingDirection, halfway));
                } else {
                    refraction = eta * eta * FresnelNonPolarized_T(dot(-incomingDirection, -halfway), bounceMaterial.n1, bounceMaterial.n2) *
                                D_wi(-incomingDirection, -halfway, alpha) * max(0.0, -dot(-outgoingDirection, -halfway)) *
                                1.0 / square(dot(-incomingDirection, -halfway) + eta * dot(-outgoingDirection, -halfway));
                }
            #else
                if(incomingOutside) {
                    refraction = 1.0 / pi;
                } else {
                    refraction = 1.0 / pi;
                }
            #endif

            return refraction;
        }
    }

    float EstimateDielectric(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounce, in float wavelength, in Material bounceMaterial) {
        float hr = 1.0 + InvC1(0.999);

        vec2 alpha = vec2(bounceMaterial.roughness);

        #ifdef CELESTIAL_PATH_ROUGHNESS_BOUNCE_INCREASE
            if(bounce > 1) {
                alpha = 0.01;
                alpha *= 15.0;
                alpha = saturate(alpha);
            }
        #endif

        bool d1_outside = incomingDirection.z < 0.0;
        bool d2_outside = outgoingDirection.z > 0.0;

        if (!d1_outside) { hr = -hr; }

        //if (d1_outside != d2_outside) { return 0.0; }

        int scatteringOrder = 0;
        float throughput = 1.0;
        float estimate = 0.0;
        while(scatteringOrder <= MAX_BSDF_SCATTERING_EVENTS) {
            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            float U = RandNextF();
            hr = d1_outside ? SampleHeight(incomingDirection, alpha, hr, U) : -SampleHeight(-incomingDirection, alpha, -hr, U);

            if(hr == 1e35 || hr == -1e35) {
                break;
            }

            scatteringOrder++;

            float phase = EstimateDielectricPhase(-incomingDirection, outgoingDirection, alpha, d1_outside, d2_outside, wavelength, bounceMaterial);
            if(isnan(phase)) {
                phase = 1.0;
            }
            float shadowing = d2_outside ? G_1(outgoingDirection, alpha, hr) : G_1(-outgoingDirection, alpha, -hr);
            estimate += throughput * phase * shadowing;

            if(isnan(throughput)) {
                throughput = 0.0;
            }

            incomingDirection = SampleDielectricPhaseFunction(-incomingDirection, d1_outside, throughput, wavelength, bounceMaterial);

            if(hr != hr || incomingDirection.z != incomingDirection.z) {
                return 0.0;
            }
        }
        if(isnan(estimate)) {
            estimate = 0.0;
        }

        return estimate / outgoingDirection.z;
    }
#endif