#if !defined BSDF_OPAQUE
#define BSDF_OPAQUE
    /*
        Calculate the opaque dielectric multiple scattering BSDF.

        This BSDF model handles an opaque dielectric material. All light that enters leaves on the side it entered.
        Most materials fall under this BSDF.

        Basically, this is a simple specular+diffuse model.
    */
    vec3 SampleOpaqueDielectricFacetBSDF(in vec3 direction, in vec3 normal, inout float estimate, inout float throughput, in float wavelength, in Material bounceMaterial, inout bool isSpecular) {
        complexFloat n1 = bounceMaterial.n1;
        complexFloat n2 = bounceMaterial.n2;

        float fresnel_R = FresnelNonPolarized_R(dot(normal, direction), n1, n2);
        float fresnel_T = FresnelNonPolarized_T(dot(normal, direction), n1, n2);

        if(bounceMaterial.thinFilm) {
            fresnel_R = FresnelThinFilmInterferenceReflected(dot(normal, direction), bounceMaterial.thickness, wavelength, n1, bounceMaterial.filmIOR, n2);
            fresnel_T = FresnelThinFilmInterferenceTransmitted(dot(normal, direction), bounceMaterial.thickness, wavelength, n1, bounceMaterial.filmIOR, n2);
        }

        float totalLum = bounceMaterial.albedo * fresnel_T + fresnel_R;
        float specBounceProbability = fresnel_R / totalLum;
        bool specularBounce = specBounceProbability > RandNextF();

        if(specularBounce) {
            isSpecular = true;
            throughput /= specBounceProbability;

            throughput *= fresnel_R;

            return reflect(-direction, normal);
        } else {
            isSpecular = false;
            throughput /= 1.0 - specBounceProbability;

            float energyConservationFactor = 1.0;

            if(bounceMaterial.thinFilm) {
                energyConservationFactor = HemisphericalAlbedoIntegration(bounceMaterial, wavelength);
            } else {
                energyConservationFactor -= HemisphericalAlbedo(bounceMaterial.n2.r / bounceMaterial.n1.r);
            }

            throughput *= fresnel_T;

            throughput /= energyConservationFactor;

            vec3 newDirection = GenerateCosineVector(normal, RandNext2F());

            throughput *= bounceMaterial.albedo * bounceMaterial.ao;

            estimate += throughput * bounceMaterial.emission;

            float fresnelL_T = FresnelNonPolarized_T(dot(normal, newDirection), bounceMaterial.n1, bounceMaterial.n2);
            if(bounceMaterial.thinFilm) {
                fresnelL_T = FresnelThinFilmInterferenceTransmitted(dot(normal, newDirection), bounceMaterial.thickness, wavelength, n1, bounceMaterial.filmIOR, n2);
            }

            throughput *= fresnelL_T;

            if(isnan(throughput)) throughput = 1.0;

            return newDirection;
        }
    }

    vec3 SampleOpaqueDielectricFacetBSDF(in vec3 direction, in vec3 normal, inout float estimate, inout float throughput, in float wavelength, in Material bounceMaterial) {
        bool isSpecular;
        return SampleOpaqueDielectricFacetBSDF(direction, normal, estimate, throughput, wavelength, bounceMaterial, isSpecular);
    }

    vec3 SampleOpaqueDielectric(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in Material bounceMaterial) {
        int scatteringOrder = 0;
        float hr = 1.0 + InvC1(0.999);
        while(scatteringOrder <= MAX_BSDF_SCATTERING_EVENTS) {
            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            float U = RandNextF();

            vec3 wm = SampleD_wi(-direction, RandNext2F(), vec2(bounceMaterial.roughness));

            hr = SampleHeight(direction, vec2(bounceMaterial.roughness), hr, U);

            if(hr == 1e35) {
                break;
            }

            direction = SampleOpaqueDielectricFacetBSDF(-direction, wm, estimate, throughput, wavelength, bounceMaterial);

            if(hr != hr || direction.z != direction.z) {
                return vec3(0, 0, 1);
            }
            scatteringOrder++;
        }

        if(isnan(throughput)) {
            throughput = 0.0;
        }

        if(scatteringOrder > MAX_BSDF_SCATTERING_EVENTS) {
            throughput = 0.0;
        }

        return direction;
    }

    float EstimateOpaqueDielectricFacetBSDF(in vec3 incomingDirection, in vec3 outgoingDirection, in vec3 normal, in float wavelength, in Material bounceMaterial, in int bounce) {
        vec3 halfway = normalize(incomingDirection + outgoingDirection);

        float vDotH = dot(incomingDirection, halfway);

        float energyConservationFactor = 1.0;

        if(bounceMaterial.thinFilm) {
            energyConservationFactor = HemisphericalAlbedoIntegration(bounceMaterial, wavelength);
        } else {
            energyConservationFactor -= HemisphericalAlbedo(bounceMaterial.n2.r / bounceMaterial.n1.r);
        }

        float diffuse  = bounceMaterial.albedo / pi;
              diffuse *= saturate(dot(outgoingDirection, normal));
              diffuse *= 1.0 - FresnelNonPolarized_R(dot(incomingDirection, normal), bounceMaterial.n1, bounceMaterial.n2);
              diffuse *= 1.0 - FresnelNonPolarized_R(dot(outgoingDirection, normal), bounceMaterial.n1, bounceMaterial.n2);
              diffuse /= energyConservationFactor;

        float numerator = D_wi(incomingDirection, halfway, vec2(bounceMaterial.roughness)) * FresnelNonPolarized_R(dot(incomingDirection, halfway), bounceMaterial.n1, bounceMaterial.n2);
        float denominator = 4.0 * vDotH;
        float specular = numerator / denominator;

        return specular + diffuse;
    }

    float EstimateOpaqueDielectric(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounce, in float wavelength, in Material bounceMaterial) {
        float hr = 1.0 + InvC1(0.999);

        vec2 alpha = vec2(bounceMaterial.roughness);

        #ifdef CELESTIAL_PATH_ROUGHNESS_BOUNCE_INCREASE
            if(bounce > 1) {
                alpha = 0.01;
                alpha *= 15.0;
                alpha = saturate(alpha);
            }
        #endif

        bool d1_outside = incomingDirection.z < 0.0;
        bool d2_outside = outgoingDirection.z > 0.0;

        if (d1_outside != d2_outside) { return 0.0; }

        int scatteringOrder = 0;
        float throughput = 1.0;
        float estimate = 0.0;
        while(scatteringOrder <= MAX_BSDF_SCATTERING_EVENTS) {
            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            float U = RandNextF();
            hr = SampleHeight(incomingDirection, alpha, hr, U);

            if(hr == 1e35) {
                break;
            }

            scatteringOrder++;

            vec3 normal = SampleD_wi(-incomingDirection, RandNext2F(), alpha);

            float phase = EstimateOpaqueDielectricFacetBSDF(-incomingDirection, outgoingDirection, normal, wavelength, bounceMaterial, bounce);
            if(isnan(phase)) {
                phase = 1.0;
            }
            estimate += throughput * phase * G_1(outgoingDirection, alpha, hr);

            if(isnan(throughput)) {
                throughput = 0.0;
            }

            float tmp;
            incomingDirection = SampleOpaqueDielectricFacetBSDF(-incomingDirection, normal, tmp, throughput, wavelength, bounceMaterial);

            if(hr != hr || incomingDirection.z != incomingDirection.z) {
                return 0.0;
            }
        }
        if(isnan(estimate)) {
            estimate = 0.0;
        }

        return estimate / outgoingDirection.z;
    }
#endif