#if !defined BSDF_OPAQUE_V
#define BSDF_OPAQUE_V
    /*
        Calculate the V-Cavity opaque dielectric multiple scattering BSDF.

        This BSDF model handles an opaque dielectric material with V shaped microfacets. All light that enters leaves on the 
        side it entered.
        Only a few materials in the real world are properly modelled by this.
    */

    vec3 SampleOpaqueDielectric_VCavity(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in Material bounceMaterial) {
        vec2 U = RandNext2F();

        vec2 alpha = vec2(bounceMaterial.roughness);

        vec3 microsurfacePlaneNormal[2];
        vec3 sampledNormal = SampleD(U, alpha);
        for (int i = 0; i < 2; i++) {
            //Calculate the microfacet normal for each plane
            float angle = float(i) * (tau / float(2));

            microsurfacePlaneNormal[i] = GetRotationMatrix(vec3(0.0, 0.0, 1.0), angle) * sampledNormal;
        }

        float height = -tan(acos(sampledNormal.z));

        vec3 rayPosition = vec3(sampledNormal.xy == vec2(0.0) ? vec2(0.0) : normalize(sampledNormal.xy) * (2.0 * RandNextF() - 1.0), 0.0);

        int scatteringOrder = 0;
        while(scatteringOrder <= MAX_BSDF_SCATTERING_EVENTS) {
            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            vec3 normal;
            vec3 point = rayPosition + direction * 3.4e38;

            bool hit = false;
            for (int i = 0; i < 2; ++i) {
                //Calculate the normals for the whole cavity, as well as update the hit point on the cavity, and whether we hit anything or not
                float t = 3.4e38;
                int p = RayPlaneIntersection(
                    rayPosition, direction,
                    vec3(0.0, 0.0, height), microsurfacePlaneNormal[i],
                    t
                );

                if(p != 0) {
                    vec3 currentPoint = rayPosition + direction * t;
                    if(currentPoint.z <= 0.0 && currentPoint.z >= height) {
                        if(distance(rayPosition, currentPoint) < distance(rayPosition, point)) {
                            normal = microsurfacePlaneNormal[i];
                            point = currentPoint;
                        }
                        hit = true;
                    }
                }
            }

            if(height == 1e35) {
                break;
            }

            scatteringOrder++;

            if(hit) {
                rayPosition = point;

                direction = SampleOpaqueDielectricFacetBSDF(-direction, normal, throughput, throughput, wavelength, bounceMaterial);
            } else {
                break;
            }

            if(height != height || direction.z != direction.z) {
                return vec3(0.0, 0.0, 1.0);
            }
        }

        if(scatteringOrder > MAX_BSDF_SCATTERING_EVENTS) {
            throughput = 0.0;
        }

        if(throughput != throughput) {
            throughput = 0.0;
        }

        return direction;
    }

    float VCavity_Shadowng(in vec3 incomingDirection, in vec3 outgoingDirection, in vec2 alpha) {
        vec2 U = RandNext2F();

        vec3 microsurfacePlaneNormal[2];
        vec3 sampledNormal = SampleD(U, alpha);
        for (int i = 0; i < 2; i++) {
            //Calculate the microfacet normal for each plane
            float angle = float(i) * (tau / float(2));

            microsurfacePlaneNormal[i] = GetRotationMatrix(vec3(0.0, 0.0, 1.0), angle) * sampledNormal;
        }

        float height = -tan(acos(sampledNormal.z));

        vec3 rayPosition = vec3(sampledNormal.xy == vec2(0.0) ? vec2(0.0) : normalize(sampledNormal.xy) * (2.0 * RandNextF() - 1.0), 0.0);

        vec3 point = rayPosition + incomingDirection * 3.4e38;

        bool hit = false;
        for (int i = 0; i < 2; ++i) {
            //Update the hit point on the cavity, and whether we hit anything or not
            float t = 3.4e38;
            int p = RayPlaneIntersection(
                rayPosition, vec3(0, 0, -1),
                vec3(0.0, 0.0, height), microsurfacePlaneNormal[i],
                t
            );

            if(p != 0) {
                vec3 currentPoint = rayPosition + vec3(0, 0, -1) * t;
                if(currentPoint.z <= 0.0 && currentPoint.z >= height) {
                    if(distance(rayPosition, currentPoint) < distance(rayPosition, point)) {
                        point = currentPoint;
                    }
                    hit = true;
                }
            }
        }

        bool inShadow = false;
        bool inView = true;

        rayPosition = point;

        for (int i = 0; i < 2; ++i) {
            //Calculate the visibility of the facet
            float t;
            int p = RayPlaneIntersection(
                rayPosition, incomingDirection,
                vec3(0.0, 0.0, height), microsurfacePlaneNormal[i],
                t
            );

            if(p != 0) {
                vec3 currentPoint = rayPosition + incomingDirection * t;
                if(currentPoint.z <= 0.0 && currentPoint.z >= height) {
                    inView = false;
                }
            }
        }

        for (int i = 0; i < 2; ++i) {
            //Calculate the shadowing of the facet
            float t;
            int p = RayPlaneIntersection(
                rayPosition, outgoingDirection,
                vec3(0.0, 0.0, height), microsurfacePlaneNormal[i],
                t
            );

            if(p != 0) {
                vec3 currentPoint = rayPosition + outgoingDirection * t;
                if(currentPoint.z <= 0.0 && currentPoint.z >= height) {
                    inShadow = true;
                }
            }
        }
        
        float shadowing = 0.0;
        if(!inShadow) {
            shadowing = 1.0;
        } else if(!inView) {
            shadowing = 0.0;
        }

        return shadowing;
    }

    float EstimateOpaqueDielectric_VCavity(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounce, in float wavelength, in Material bounceMaterial) {
        vec2 U = RandNext2F();

        vec2 alpha = vec2(bounceMaterial.roughness);

        vec3 microsurfacePlaneNormal[2];
        vec3 sampledNormal = SampleD(U, alpha);
        for (int i = 0; i < 2; i++) {
            //Calculate the microfacet normal for each plane
            float angle = float(i) * (tau / float(2));

            microsurfacePlaneNormal[i] = GetRotationMatrix(vec3(0.0, 0.0, 1.0), angle) * sampledNormal;
        }

        float height = -tan(acos(sampledNormal.z));

        vec3 rayPosition = vec3(sampledNormal.xy == vec2(0.0) ? vec2(0.0) : normalize(sampledNormal.xy) * (2.0 * RandNextF() - 1.0), 0.0);

        float estimate = 0.0;
        float throughput = 1.0;

        bool d1_outside = incomingDirection.z < 0.0;
        bool d2_outside = outgoingDirection.z > 0.0;

        if (d1_outside != d2_outside) { return 0.0; }

        vec3 oldIncomingDirection = -incomingDirection;
        int scatteringOrder = 0;
        while(scatteringOrder <= MAX_BSDF_SCATTERING_EVENTS) {
            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            vec3 normal;
            vec3 point = rayPosition + incomingDirection * 3.4e38;

            bool hit = false;
            for (int i = 0; i < 2; ++i) {
                //Calculate the normals for the whole cavity, as well as update the hit point on the cavity, and whether we hit anything or not
                float t = 3.4e38;
                int p = RayPlaneIntersection(
                    rayPosition, incomingDirection,
                    vec3(0.0, 0.0, height), microsurfacePlaneNormal[i],
                    t
                );

                if(p != 0) {
                    vec3 currentPoint = rayPosition + incomingDirection * t;
                    if(currentPoint.z <= 0.0 && currentPoint.z >= height) {
                        if(distance(rayPosition, currentPoint) < distance(rayPosition, point)) {
                            normal = microsurfacePlaneNormal[i];
                            point = currentPoint;
                        }
                        hit = true;
                    }
                }
            }

            if(height == 1e35) {
                break;
            }

            scatteringOrder++;

            if(hit) {
                rayPosition = point;

                bool inShadow = false;
                for (int i = 0; i < 2; ++i) {
                    //Calculate the shadowing
                    float t;
                    int p = RayPlaneIntersection(
                        rayPosition, outgoingDirection,
                        vec3(0.0, 0.0, height), microsurfacePlaneNormal[i],
                        t
                    );

                    if(p != 0) {
                        vec3 currentPoint = rayPosition + outgoingDirection * t;
                        if(currentPoint.z <= 0.0 && currentPoint.z >= height) {
                            inShadow = true;
                        }
                    }
                }

                if(alpha.x == 0.0 && alpha.y == 0.0) {
                    //This makes sure the bsdf does not break when roughness is zero.
                    inShadow = false;
                    normal = vec3(0.0, 0.0, 1.0);
                }

                if(!inShadow && !bounceMaterial.isMetal) {
                    vec3 halfway = normalize(incomingDirection + outgoingDirection);

                    float vDotH = dot(incomingDirection, halfway);

                    float energyConservationFactor = 1.0;

                    if(bounceMaterial.thinFilm) {
                        energyConservationFactor = HemisphericalAlbedoIntegration(bounceMaterial, wavelength);
                    } else {
                        energyConservationFactor -= HemisphericalAlbedo(bounceMaterial.n2.r / bounceMaterial.n1.r);
                    }

                    float diffuse  = bounceMaterial.albedo / pi;
                          diffuse *= saturate(dot(outgoingDirection, normal));
                          diffuse *= 1.0 - FresnelNonPolarized_R(dot(incomingDirection, normal), bounceMaterial.n1, bounceMaterial.n2);
                          diffuse *= 1.0 - FresnelNonPolarized_R(dot(outgoingDirection, normal), bounceMaterial.n1, bounceMaterial.n2);
                          diffuse /= energyConservationFactor;

                    estimate += throughput * diffuse;
                }

                float tmp;
                incomingDirection = SampleOpaqueDielectricFacetBSDF(-incomingDirection, normal, tmp, throughput, wavelength, bounceMaterial);
            } else {
                break;
            }

            if(height != height || incomingDirection.z != incomingDirection.z) {
                return 0.0;
            }
        }

        vec3 halfway = normalize(oldIncomingDirection + outgoingDirection);
        float F = FresnelNonPolarized_R(dot(oldIncomingDirection, halfway), bounceMaterial.n1, bounceMaterial.n2);
        float G = VCavity_Shadowng(oldIncomingDirection, outgoingDirection, alpha);
        float numerator = D(halfway, alpha) * F * G;
        float denominator = 4.0 * outgoingDirection.z * oldIncomingDirection.z;
        float specular = numerator / denominator;

        return (estimate / outgoingDirection.z) + max(specular, 0.0);
    }
#endif