#if !defined MULTIPLE_SCATTERING_BSDF_LIB
#define MULTIPLE_SCATTERING_BSDF_LIB
    float G_1(in vec3 wi, in vec2 alpha, in float h0) {
        if(wi.xy == vec2(0.0)) {
            return 1.0;
        }
        if(wi.z <= 0.0) {
            return 0.0;
        }

        float C1_h0 = C1(h0);
        float lambda = Smith_Lambda(wi, alpha);
        float value = pow(C1_h0, lambda);

        return value;
    }

    float G_1(in vec3 wi, in vec2 alpha) {
        if(wi.xy == vec2(0.0)) {
            return 1.0;
        }
        if(wi.z <= 0.0) {
            return 0.0;
        }

        float lambda = Smith_Lambda(wi, alpha);
        float value = 1.0 / (1.0 + lambda);

        return value;
    }

    float SampleHeight(in vec3 wr, in vec2 alpha, in float hr, in float U) {
        if(wr.z > 0.9999) {
            return 1e35;
        }
        if(wr.z < -0.9999) {
            float value = InvC1(U * C1(hr));
            return value;
        }
        if(abs(wr.z) < 0.0001) {
            return hr;
        }

        float G1 = G_1(wr, alpha, hr);

        if(U > 1.0 - G1) {
            return 1e35;
        }

        float h = InvC1(
            C1(hr) / pow(1.0 - U, 1.0 / Smith_Lambda(wr, alpha))
        );

        return h;
    }

    float HemisphericalAlbedo(in float n) {
        float n2 = square(n);
        float T_1 = (4.0 * (2.0 * n + 1.0)) / (3.0 * square(n + 1.0));
        float T_2 = ((4.0 * cube(n) * (n2 + 2.0 * n - 1.0)) / (square(n2 + 1.0) * (n2 - 1.0))) - 
                ((2.0 * n2 * (n2 + 1.0) * log(n)) / square(n2 - 1.0)) +
                ((2.0 * n2 * square(n2 - 1.0) * log((n * (n+1.0)) / (n-1.0))) / cube(n2 + 1.0));
        return saturate(1.0 - 0.5 * (T_1 + T_2));
    }

    float HemisphericalAlbedoIntegration(in Material bounceMaterial, in float wavelength) {
        float energyConservationFactor = 1.0;
        uint N = 32u; //This amount of directions gives low bias while still remaining relatively fast.
        for(uint i = 0u; i < N; ++i) {
            //Calculate the energy conservation factor via brute force integration.
            //This is required for thin film interference due to the lack of a closed form solution.
            float nDotV = 1.0 - RandNext2F().x; //This is equivalent to nDotV here.
            float fresnel = FresnelThinFilmInterferenceReflected(
                                nDotV, 
                                bounceMaterial.thickness, 
                                wavelength, 
                                bounceMaterial.n1, 
                                bounceMaterial.filmIOR, 
                                bounceMaterial.n2
                            );
            energyConservationFactor -= fresnel * (2.0 * nDotV) / float(N);
        }

        return energyConservationFactor;
    }

    vec3 wt(in vec3 wp) {
        return normalize(vec3(-wp.x, -wp.y, 0.0));
    }

    float a_p(vec3 wm, vec3 wi) {
        return dot(wi, wm) / wm.z;
    }

    float a_t(vec3 wm, vec3 wi) {
        return (dot(wi, wt(wm)) * sqrt(1.0 - square(wm.z))) / wm.z;
    }

    float lambda_p(vec3 wm, vec3 wi) {
        return a_p(wm, wi) / (a_p(wm, wi) + a_t(wm, wi));
    }

    float lambda_t(vec3 wm, vec3 wi) {
        return a_t(wm, wi) / (a_p(wm, wi) + a_t(wm, wi));
    }

    float Heaviside(in float x) {
        return float(x > 0.0);
    }

    float G1_MN(vec3 wi, vec3 wp, vec3 wm) {
        if (wi.z < 0.0) { return 0.0; }

        return Heaviside(dot(wi, wm)) * min(1.0, wi.z / (a_p(wp, wi) + a_t(wp, wi)));
    }

    #include "BSDFNodes/opaque.glsl"
    #include "BSDFNodes/vcavity_opaque.glsl"
    #include "BSDFNodes/dielectric.glsl"
    #include "BSDFNodes/conductor.glsl"
    #include "BSDFNodes/lambertian.glsl"

    vec3 SampleOpaqueSelector(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in Material bounceMaterial) {
        #if MICROSURFACE_PROFILE == 0
            return SampleOpaqueDielectric(direction, estimate, throughput, wavelength, bounceMaterial);
        #else
            return SampleOpaqueDielectric_VCavity(direction, estimate, throughput, wavelength, bounceMaterial);
        #endif
    }

    float EstimateOpaqueSelector(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounce, in float wavelength, in Material bounceMaterial) {
        #if MICROSURFACE_PROFILE == 0
            return EstimateOpaqueDielectric(incomingDirection, outgoingDirection, bounce, wavelength, bounceMaterial);
        #else
            return EstimateOpaqueDielectric_VCavity(incomingDirection, outgoingDirection, bounce, wavelength, bounceMaterial);
        #endif
    }

    #include "BSDFNodes/normalMapping.glsl"
    #include "BSDFNodes/clearcoat.glsl"
#endif