struct LensInterface {
	float axisPosition; // Position of interface along the lens axis, as measured at its center.
	float axisRadius;   // Radius of the interface around the lens axis.

	int shapeType; // Shape type of interface. 0 is flat, 1 is spherical.
	float curvatureRadius; // Radius of curvature, for spherical interfaces.

	bool hasThinCoating; // If this interface has some kind of coating with negligible thickness. Primarily for thin-film interference.
	float coatIor;       // Index of Refraction of the coating.
	float coatThickness; // Thickness of the coating.
};

const int num_interfaces = 16;
LensInterface[num_interfaces] interfaces = LensInterface[num_interfaces](
	LensInterface(10.00, 1e35, 1,  164.13, false, 1.0, 0.0),
	LensInterface( 0.23, 1e35, 1,  559.20, false, 1.0, 0.0),
	LensInterface(11.45, 1e35, 1,  100.12, false, 1.0, 0.0),
	LensInterface( 0.23, 1e35, 1,  213.54, false, 1.0, 0.0),
	LensInterface(22.95, 1e35, 1,   58.04, false, 1.0, 0.0),
	LensInterface( 2.58, 1e35, 1, 2551.10, false, 1.0, 0.0),
	LensInterface(30.66, 1e35, 1,   32.39, false, 1.0, 0.0),
	LensInterface(10.00, 1e35, 0, 1.0/0.0, false, 1.0, 0.0), // aperture?
	LensInterface( 2.74, 1e35, 1, -40.042, false, 1.0, 0.0),
	LensInterface(27.92, 1e35, 1,  192.98, false, 1.0, 0.0),
	LensInterface( 0.23, 1e35, 1,  -55.53, false, 1.0, 0.0),
	LensInterface( 7.98, 1e35, 1,  192.98, false, 1.0, 0.0),
	LensInterface( 0.23, 1e35, 1, -225.30, false, 1.0, 0.0),
	LensInterface( 8.48, 1e35, 1,  175.09, false, 1.0, 0.0),
	LensInterface(40.00, 1e35, 1, -203.55, false, 1.0, 0.0),
	LensInterface( 5.00, 1e35, 0, 1.0/0.0, false, 1.0, 0.0) // sensor?
);
float[num_interfaces - 1] refractiveIndices = float[num_interfaces - 1](
	1.6751,
	1.0,
	1.6689,
	1.0,
	1.69131,
	1.6751,
	1.0,
	1.0,
	1.6992,
	1.6204,
	1.0,
	1.69131,
	1.0,
	1.6913,
	1.0
);

bool invalid = false;
for (int interfaceIndex = 0; interfaceIndex < num_interfaces; ++interfaceIndex) {
	LensInterface interface = interfaces[interfaceIndex];

	vec3 interfaceNormal;
	if (interface.shapeType == 1) {
		vec3 sphereOrigin = vec3(0.0, 0.0, interface.axisPosition - interface.curvatureRadius);

		float t0, t1;
		int intersectionType = IntersectRaySphere(
			rayPosition, rayDirection,
			sphereOrigin, interface.curvatureRadius,
			t0, t1
		);

		if (intersectionType == 0) {
			invalid = true;
			break;
		} else if (intersectionType == 1) {
			if (interface.curvatureRadius > 0.0) {
				invalid = true;
				break;
			}

			rayPosition += rayDirection * t0;
		} else if (intersectionType == 2) {
			if (interface.curvatureRadius < 0.0) {
				rayPosition += rayDirection * t1;
			} else {
				rayPosition += rayDirection * t0;
			}
		} else {
			// This should be unreachable.
			invalid = true;
			break;
		}

		interfaceNormal = normalize(rayPosition - sphereOrigin);
	} else if (interface.shapeType == 0) {
		interfaceNormal = vec3(0.0, 0.0, 1.0);

		float t;
		int intersectionType = IntersectRayPlane(
			rayPosition, rayDirection,
			vec3(0.0, 0.0, interface.axisPosition), interfaceNormal,
			t
		);

		if (intersectionType == 0) {
			invalid = true;
			break;
		} else if (intersectionType == 1) {
			rayPosition += rayDirection * t;
		} else if (intersectionType == 2) {
			invalid = true;
			break;
		} else {
			// This should be unreachable.
			invalid = true;
			break;
		}
	}

	float n1 = interfaceIndex == 0                  ? 1.0 : refractiveIndices[interfaceIndex - 1];
	float n2 = interfaceIndex == num_interfaces - 1 ? 1.0 : refractiveIndices[interfaceIndex];

	rayDirection = refract(rayDirection, interfaceNormal, n2 / n1);
	// Check for Total Internal Reflection
	if (rayDirection == vec3(0.0)) {
		invalid = true;
		break;
	}

	if (interface.hasThinCoating) {
		// Fresnel with thin-film interference
	} else {
		// Normal fresnel
	}
}