#if !defined PATH_ATMOSPHERE
#define PATH_ATMOSPHERE
    const float kilometer = 1000.0;
    const float ozoneHeight = 10.0;
    const float airNumberDensity   = 2.5035422e25;
    const float ozoneNumberDensity = airNumberDensity * exp(-25e3 / 8e3) * (134.628/48.0) * 1e-6;
    const float planetRadius       = 6371e3;
    const float atmosphereHeight   = 110e3;

    const vec2 scaleHeights = vec2(8.0, 1.2) * kilometer;

    const float mieG = 0.82;

    const vec2 inverseScaleHeights   = 1.0 / (scaleHeights);
    const vec2 scaledPlanetRadius    = planetRadius * inverseScaleHeights;
    const float atmosphereRadius     = planetRadius + atmosphereHeight;
    const float atmosphereRadiusSquared = atmosphereRadius*atmosphereRadius; 
    const float atmosphereLowerLimit = planetRadius + -1.0;
    const float atmosphereLowerLimitSquared = atmosphereLowerLimit*atmosphereLowerLimit;

    const float aerosolScatteringAlbedo = 0.90;
    const float rainbowScatteringAlbedo = 0.99;

    const float densitySeaLevel = 1.225;
    const float gravity = 9.80665;
    const float rAir = 287.053;
    const float gammaAir = 1.4;

    float RayleighDensity(in float h, in float cvza) {
        return exp(-h * inverseScaleHeights.x + scaledPlanetRadius.x);
    }

    float AerosolDensity(in float h) {
        return exp(-h * inverseScaleHeights.y + scaledPlanetRadius.y);
    }

    float OzoneDensity(in float altitudeKm) {
        float o1 = 25.0 *     exp(( 0.0 - altitudeKm) /   8.0) * 0.5;
        float o2 = 30.0 * pow(exp((18.0 - altitudeKm) /  80.0), altitudeKm - 18.0);
        float o3 = 75.0 * pow(exp((25.3 - altitudeKm) /  35.0), altitudeKm - 25.3);
        float o4 = 50.0 * pow(exp((30.0 - altitudeKm) / 150.0), altitudeKm - 30.0);
        return (o1 + o2 + o3 + o4) / 134.628;
    }

    vec3 CalculateAtmosphereDensity(in float h, in float cvza) {
        vec2 rm;
        rm.x = RayleighDensity(h, cvza);
        rm.y = AerosolDensity(h);

        float altitudeKm = (h - planetRadius) / kilometer;

        float ozone = OzoneDensity(altitudeKm);

        return vec3(rm, ozone);
    }

    const float cloudsAltitude = 600.0;
    const float cloudsThickness = cloudsAltitude * 1.2;
    const float cloudsMaxAltitude = cloudsAltitude + cloudsThickness;

    const float cloudsDensityConstant = 0.0;

    float SampleNullDistance(
        vec3  rayPosition,
        vec3  rayVector,
        vec3 baseAttenuationCoefficients,
        out float localMajorant
    ) {
        const vec3 planetPosition = vec3(0.0);

        float tScale = length(rayVector);

        float altitude = distance(rayPosition, planetPosition) - planetRadius;

        if (dot(rayPosition - planetPosition, rayVector) < 0.0) {
            // Ray is (locally) moving downwards
            // Find minimum altitude along the ray

            float tMinAltitude = dot(planetPosition - rayPosition, rayVector) / dot(rayVector, rayVector);
            vec3  pMinAltitude = rayPosition + rayVector * tMinAltitude;
            float minAltitude = distance(planetPosition, pMinAltitude) - planetRadius;

            minAltitude = max(minAltitude, 0.0);
            pMinAltitude = planetPosition + vec3(0.0, planetRadius + minAltitude, 0.0);

            localMajorant  = baseAttenuationCoefficients.x * CalculateAtmosphereDensity(planetRadius + minAltitude, dot(rayPosition, rayVector) / minAltitude).x;
            localMajorant += baseAttenuationCoefficients.y * CalculateAtmosphereDensity(planetRadius + minAltitude, dot(rayPosition, rayVector) / minAltitude).y;

            if (altitude <= cloudsAltitude) {
                // Below cloud layer
                localMajorant += baseAttenuationCoefficients.z * CalculateAtmosphereDensity(planetRadius + minAltitude, dot(rayPosition, rayVector) / minAltitude).z;
            } else if (altitude < cloudsMaxAltitude) {
                // Inside cloud layer
                localMajorant += baseAttenuationCoefficients.z * CalculateAtmosphereDensity(planetRadius + minAltitude, dot(rayPosition, rayVector) / minAltitude).z;
            } else if (altitude < 25e3) {
                // Below ozone peak
                localMajorant += max(
                    baseAttenuationCoefficients.z * CalculateAtmosphereDensity(planetRadius + minAltitude, dot(rayPosition, rayVector) / minAltitude).z,
                    baseAttenuationCoefficients.z * CalculateAtmosphereDensity(altitude + planetRadius, dot(rayPosition, rayVector) / (atmosphereLowerLimit)).z
                );
            } else {
                // Above ozone peak
                localMajorant += max(
                    baseAttenuationCoefficients.z * CalculateAtmosphereDensity(planetRadius + minAltitude, dot(rayPosition, rayVector) / minAltitude).z,
                    baseAttenuationCoefficients.z * CalculateAtmosphereDensity(atmosphereLowerLimit + 25e3, dot(rayPosition, rayVector) / (atmosphereLowerLimit + 25e3)).z
                );
            }

            float tSampled = -log(RandNextF()) / (tScale * localMajorant);

            if (altitude >= cloudsMaxAltitude) {
                float t0, t1;
                if (LineSphereIntersect(rayPosition, rayVector, planetPosition, planetRadius + cloudsMaxAltitude, t0, t1) != 0 && t0 < tSampled) {
                    tSampled = t0 + -log(RandNextF()) / (tScale * localMajorant);
                }
            }

            if (altitude <= cloudsAltitude) {
                float t0, t1;
                LineSphereIntersect(rayPosition, rayVector, planetPosition, planetRadius + cloudsAltitude, t0, t1);
                if (t1 < tSampled) {
                    tSampled = t1 + -log(RandNextF()) / (tScale * localMajorant);
                }
            }

            return tSampled;
        } else {
            // Ray is (locally) moving upwards (or horizontally)
            localMajorant  = baseAttenuationCoefficients.x * CalculateAtmosphereDensity(altitude + planetRadius, dot(rayPosition, rayVector) / atmosphereLowerLimit).x;
            localMajorant += baseAttenuationCoefficients.y * CalculateAtmosphereDensity(altitude + planetRadius, dot(rayPosition, rayVector) / atmosphereLowerLimit).y;

            if (altitude <= cloudsAltitude) {
                // Below cloud layer
                localMajorant += baseAttenuationCoefficients.z * CalculateAtmosphereDensity(altitude + planetRadius, dot(rayPosition, rayVector) / (atmosphereLowerLimit)).z;
            } else if (altitude < cloudsMaxAltitude) {
                // Inside cloud layer
                localMajorant += baseAttenuationCoefficients.z * CalculateAtmosphereDensity(altitude + planetRadius, dot(rayPosition, rayVector) / (atmosphereLowerLimit)).z;
            } else if (altitude < 25e3) {
                // Below ozone peak
                localMajorant += baseAttenuationCoefficients.z * CalculateAtmosphereDensity(atmosphereLowerLimit+25e3, dot(rayPosition, rayVector) / (atmosphereLowerLimit+25e3)).z;
            } else {
                // Above ozone peak
                localMajorant += baseAttenuationCoefficients.z * CalculateAtmosphereDensity(altitude + planetRadius, dot(rayPosition, rayVector) / (atmosphereLowerLimit)).z;
            }

            float tSampled = -log(RandNextF()) / (tScale * localMajorant);

            if (altitude <= cloudsAltitude) {
                float t0, t1;
                LineSphereIntersect(rayPosition, rayVector, planetPosition, planetRadius + cloudsAltitude, t0, t1);
                if (t1 < tSampled) {
                    tSampled = t1 + -log(RandNextF()) / (tScale * localMajorant);
                }
            }

            return tSampled;
        }
    }

    int FindNextInteraction(in vec3 position, in vec3 direction, in vec3 baseAttenuationCoefficients, out int component, out float volumeDistance, in float tMax, in float wavelength) {
        position = SceneSpaceFromVoxelSpace(position) + vec3(0.0, planetRadius + cameraPosition.y, 0.0);
        direction = normalize(direction);
        vec2 aid = RSI(position, direction, atmosphereRadius);

        float t = max(aid.x, 0.0);
        float t1 = min(aid.y, tMax);

        float maxAttenuationCoefficient;

        vec3 startPosition = position;
        vec3 oldDirection = direction;

        t += SampleNullDistance(
            startPosition,
            direction,
            baseAttenuationCoefficients,
            maxAttenuationCoefficient
        );

        int steps = 0;
        while (t < t1) {
            //if(steps > 10000) break; //Introduces bias, but prevents infinite loops and crashing on some devices

            position = startPosition + direction * t;

            if(length(position) <= planetRadius) {
                return 2;

                break;
            }
            if(length(position) > atmosphereRadius) {
                return 3;

                break;
            }

            float densityRayleigh = CalculateAtmosphereDensity(length(position), dot(position, direction) / length(position)).x;
            float densityMie      = CalculateAtmosphereDensity(length(position), dot(position, direction) / length(position)).y;
            float densityOzo      = CalculateAtmosphereDensity(length(position), dot(position, direction) / length(position)).z;
            if (densityOzo > 1e35) { break; }
            if (densityMie > 1e35) { break; }
            if (densityRayleigh > 1e35) { break; }

            float rayleighAttenuation = baseAttenuationCoefficients.x * densityRayleigh;
            float mieAttenuation      = baseAttenuationCoefficients.y * densityMie;
            float ozoneAttenuation    = baseAttenuationCoefficients.z * densityOzo;
            vec3 stepAttenuationCoeffcients = vec3(rayleighAttenuation, mieAttenuation, ozoneAttenuation);

            float rand = RandNextF();
            if (rand < ((rayleighAttenuation + mieAttenuation + ozoneAttenuation) / maxAttenuationCoefficient)) {
                float fraction = 0.0;
                component = 0;
                while (component < 2) {
                    fraction += stepAttenuationCoeffcients[component];
                    if (rand * maxAttenuationCoefficient < fraction) { break; }
                    ++component;
                }
                volumeDistance = t;
                return 1;
            }

            t += SampleNullDistance(
                position,
                direction,
                baseAttenuationCoefficients,
                maxAttenuationCoefficient
            );

            ++steps;
        }

        return 0;
    }

    float EstimateTransmittance(in vec3 position, vec3 direction, in vec3 baseAttenuationCoefficients) {
        position = SceneSpaceFromVoxelSpace(position) + vec3(0.0, planetRadius + cameraPosition.y, 0.0);
        vec2 aid = RSI(position, direction, atmosphereRadius);
        if (aid.y < 0.0) { return 0.0; }
        vec2 pid = RSI(position, direction, atmosphereLowerLimit);
        bool planetIntersected = pid.x > 0.0;
        if(planetIntersected) { return 0.0; }
        bool atmosphereIntersected = aid.y >= 0.0;

        float tMax = (planetIntersected && pid.x > 0.0) ? pid.x : aid.y;

        float maxAttenuationCoefficient;

        vec3 previousPosition = position;

        float transmittance = 1.0;
        float t = 0.0;
        int steps = 0;
        while (t < tMax) {
            //if(steps > 10000) break; //Introduces bias, but prevents infinite loops and crashing on some devices
            float stepSize = SampleNullDistance(
                position,
                direction,
                baseAttenuationCoefficients,
                maxAttenuationCoefficient
            );

            t += stepSize;

            position = previousPosition + direction * stepSize;

            if (length(position) <= planetRadius) {
                transmittance *= 0.0;
                return transmittance;
                break; 
            }
            if (length(position) > atmosphereRadius) { break; }

            float densityRayleigh = CalculateAtmosphereDensity(length(position), dot(position, direction) / length(position)).x;
            float densityMie      = CalculateAtmosphereDensity(length(position), dot(position, direction) / length(position)).y;
            float densityOzo      = CalculateAtmosphereDensity(length(position), dot(position, direction) / length(position)).z;
            if (densityOzo > 1e35) { break; }
            if (densityMie > 1e35) { break; }
            if (densityRayleigh > 1e35) { break; }

            float rayleighAttenuation = baseAttenuationCoefficients.x * densityRayleigh;
            float mieAttenuation      = baseAttenuationCoefficients.y * densityMie;
            float ozoneAttenuation    = baseAttenuationCoefficients.z * densityOzo;

            transmittance *= 1.0 - ((rayleighAttenuation + mieAttenuation + ozoneAttenuation) / maxAttenuationCoefficient);

            previousPosition = position;

            ++steps;
        }

        return saturate(transmittance);
    }

	vec3 SampleHenyeyGreenstein(float P, float g) {
		float s = 2.0 * P - 1.0;
		float u = (0.5 / g) * (1.0 + g * g - pow((1.0 - g * g) / (1.0 + g * s), 2.0));

		return GenerateUnitVector(vec2(RandNextF(), u * 0.5 + 0.5));
	}

    float HenyeyGreensteinPhase(float cosTheta, float g) {
        const float norm = 0.25/pi;

        float gg = g * g;
        return norm * ((1.0-gg) / pow(1.0+gg-2.0*g*cosTheta, 3.0/2.0));
    }

	vec3 SamplePhaseFunction() {
		return SampleHenyeyGreenstein(RandNextF(), mieG);
	}

	float SampleRayleighPhase(float s) {	
		float t0 = pow(sqrt((s*s - s) * 16.0 + 5.0) + 4.0 * s - 2.0, 1.0/3.0);

		float cosTheta = t0 - 1.0/t0;

		return cosTheta;
	}

	vec3 SampleRayleigh() {	
		float cosTheta = SampleRayleighPhase(RandNextF());

		return GenerateUnitVector(vec2(RandNextF(), cosTheta * 0.5 + 0.5));
	}

    float RayleighPhaseSimple(float cosTheta) {
        return (square(cosTheta) * 0.75 + 0.75) * rcp(pi4);
    }

    float KleinNishina(float cosTheta, float g) {
        float e = 1.0;
        for (int i = 0; i < 8; ++i) {
            float gFromE = 1.0 / e - 2.0 / log(2.0 * e + 1.0) + 1.0;
            float deriv = 4.0 / ((2.0 * e + 1.0) * square(log(2.0 * e + 1.0))) - 1.0 / square(e);
            if (abs(deriv) < 0.00000001) break;
            e = e - (gFromE - g) / deriv;
        }

        return e / (2.0 * pi * (e * (1.0 - cosTheta) + 1.0) * log(2.0 * e + 1.0));
    }

    vec3 SampleKleinNishina(float g) {
        float e = 1.0;
        for (int i = 0; i < 8; ++i) {
            float gFromE = 1.0 / e - 2.0 / log(2.0 * e + 1.0) + 1.0;
            float deriv = 4.0 / ((2.0 * e + 1.0) * square(log(2.0 * e + 1.0))) - 1.0 / square(e);
            if (abs(deriv) < 0.00000001) break;
            e = e - (gFromE - g) / deriv;
        }

        float cosTheta = (-pow(2.0 * e + 1.0, 1.0 - RandNextF()) + e + 1.0) / e;
        return GenerateUnitVector(vec2(RandNextF(), cosTheta * 0.5 + 0.5));
    }
#endif