#if !defined SPECTRAL_MATERIAL_STRUCT_LIB
#define SPECTRAL_MATERIAL_STRUCT_LIB
    struct Material {
        //Principled BSDF material array
        vec3 normal;
        vec3 flatNormal;
        vec3 tangentNormal;

        float emission;
        float albedo;
        float metalAlbedo;

        float roughness;
        float ccRoughness;
        float clearCoat;
        float ao;
        float thickness;
        float radius;

        float subsurfaceForwardScatterG;
        float subsurfaceBackwardScatterG;

        float layerWeight; //Essentially metalness, but accurately named and used in a more realistic way

        bool isMetal;
        bool thinFilm;
        bool doRefract;
        bool isLambertian;

        bool emissive;

        complexFloat n1;
        complexFloat n2;
        complexFloat clearCoatN;
        complexFloat filmIOR;
    };

    #include "/lib/fragment/bsdf/metals.glsl"
    //#include "/lib/fragment/bsdf/ior.glsl"
    uniform sampler2D colortex14;

    float f0ToIOR(in float f0, in float wavelength) {
        f0 = sqrt(f0);
        f0 *= 0.99999; // Prevents divide by 0
        return Air(wavelength * 1e-3) * ((1.0 + f0) / (1.0 - f0));
    }

    float IORTof0(in float ior, in float wavelength) {
        float sqrtf0 = (ior - Air(wavelength * 1e-3)) / (ior + Air(wavelength * 1e-3));
        return square(sqrtf0);
    }

    float GetN(in float r, in float g, in float wavelength) {
        float rSqrt = sqrt(r);
        float nMin = (1.0 - r) / (1.0 + r);
        float nMax = f0ToIOR(r, wavelength) / Air(wavelength * 1e-3);

        return mix(nMax, nMin, g);
    }

    float GetK(in float r, in float n) {
        float k2 = ((n + 1.0) * (n + 1.0) * r - (n - 1.0) * (n - 1.0)) / (1.0 - r);
        return sqrt(k2);
    }

    float GetR(in float n, in float k, in float wavelength) {
        return (square(n - Air(wavelength * 1e-3)) + square(k)) / (square(n + Air(wavelength * 1e-3)) + square(k));
    }

    float GetG(in float n, in float k, in float wavelength) {
        float r = GetR(n, k, wavelength);

        float nMin = (1.0 - r) / (1.0 + r);
        float nMax = f0ToIOR(r, wavelength) / Air(wavelength * 1e-3);

        return (nMax - n) / (nMax - nMin);
    }

    float Get3DNoise(in vec3 position) {
        float flr = floor(position.z);
        vec2 coord = (position.xy * 0.015625) + (flr * 0.265625);
        vec2 noise = texture(noisetex, coord).xy;
        return mix(noise.x, noise.y, position.z - flr);
    }

    Material hitMaterial(VoxelData voxel, in vec3 position, in float wavelength) {
        Material b;

        b.emissive = false;
        if(voxel.id == 158) {
            voxel.color = voxel.color * dot(voxel.color, vec3(1.0, 0.4, 0.2));
        }
        b.doRefract = false;

        b.clearCoat = 0.0;
        b.ccRoughness = 0.0;

        b.isMetal = voxel.specular.g > 0.9;

        b.layerWeight = 0.0;

        const float averageBrightness = 0.0878801178; //This is the average brightness of emissive objects. Any change between them shall be a multiplier, alongside this, on the brightness.

        float brightness = 0.0; //Brightness of non-emissives is technically greater than zero, but there is no reason to handle them as emissive due to how low the value would be. The accuracy gain is irrelevant due to the fact special tools, that I am not simulating, are needed to read the values.
        if(voxel.id == 18 || voxel.id == 170 || voxel.id == 172) {
            //voxel.color = BlackbodyRGB(5200.0);
            brightness = averageBrightness;
        }

        float albedo = 0.0;
        float emissionSpectra = 0.0;
        RGBToSpectrum(albedo, wavelength, voxel.color.r, voxel.color.g, voxel.color.b, 0);
        RGBToSpectrum(emissionSpectra, wavelength, voxel.color.r, voxel.color.g, voxel.color.b, 1);
        b.metalAlbedo = saturate(albedo);
        albedo = saturate(albedo * (1.0 - float(b.isMetal)));
        emissionSpectra = emissionSpectra * (1.0 - float(b.isMetal));

        b.albedo = albedo;

        b.ao = voxel.ambientOcclusion;

        b.subsurfaceForwardScatterG = 0.65;
        b.subsurfaceBackwardScatterG = -0.3;

        if(voxel.id == 17) {
            b.albedo = b.albedo * 0.1;

            voxel.specular.r = 1.0;
            voxel.specular.g = IORTof0(mix(1.345, 1.320, (wavelength - 390.0) / 441.0), wavelength);
            voxel.specular.b = 0.0;
            voxel.specular.a = 1.0;

            b.doRefract = false;
        }
        if(voxel.id == 157) {
            //voxel.specular.r = 0.0;
            voxel.specular.g = IORTof0(1.3333 + mix(-0.02, 0.04, RandNextF()), wavelength);
            //voxel.specular.b = 0.0;
            voxel.specular.a = 1.0;

            b.radius = (b.albedo * (2.0 - b.albedo));

            b.doRefract = b.radius < 0.00001 ? false : voxel.refractive;
            if(b.doRefract) {
                b.albedo = b.albedo;
            }
        }
        if(voxel.id == 158) {
            //voxel.specular.r = 1.0;
            voxel.specular.g = IORTof0(mix(1.4325, 1.4209, (wavelength - 390.0) / 441.0), wavelength);
            //voxel.specular.b = 0.0;
            voxel.specular.a = 1.0;
            if(saturate(((voxel.specular.b * 255.0)-65.0)/(255.0-65.0)) == 0.0) {
                voxel.specular.b = 1.0;
            }

            b.radius = b.albedo * saturate(((voxel.specular.b * 255.0)-65.0)/(255.0-65.0));

            b.doRefract = b.radius < min(RandNextF(), 1e-5) ? false : voxel.refractive;
            if(b.doRefract) {
                b.albedo = saturate(b.albedo);
            }
        }
        if(voxel.id == 159) {
            //voxel.specular.r = 1.0;
            voxel.specular.g = IORTof0(mix(1.4325, 1.4209, (wavelength - 390.0) / 441.0), wavelength);
            //voxel.specular.b = 0.0;
            voxel.specular.a = 1.0;

            b.radius = b.albedo * saturate(((voxel.specular.b * 255.0)-65.0)/(255.0-65.0));

            b.doRefract = b.radius < min(RandNextF(), 1e-5) ? false : voxel.refractive;
            if(b.doRefract) {
                b.albedo = saturate(b.albedo);
            }
        }
        if(voxel.id == 161) {
            voxel.specular.r = 1.0;
            voxel.specular.g = IORTof0(SellmeierEquation(wavelength), wavelength);
            voxel.specular.b = 0.0;
            voxel.specular.a = 1.0;

            b.radius = b.albedo * 2.0;

            b.doRefract = voxel.refractive;
        }

        if(voxel.id == 166) {
            voxel.specular.r = 0.0;
            voxel.specular.g = IORTof0(mix(1.3, 1.4, (wavelength - 390.0) / 441.0), wavelength);
            voxel.specular.b = 0.0;
            voxel.specular.a = 1.0;

            b.radius = b.albedo;

            b.doRefract = voxel.refractive;
        }
        if(voxel.id == 999) {
            voxel.specular.r = 0.4;
            voxel.specular.g = mix(0.07, 0.01, (wavelength - 390.0) / 441.0);
            voxel.specular.b = 0.0;
            voxel.specular.a = 1.0;

            b.doRefract = false;
        }

        b.roughness = square(1.0 - voxel.specular.r);

        float emissiveFactor = 1.0 - (voxel.specular.a * 255.0 - 254.0);
        if(voxel.specular.a < 1.0 && voxel.specular.a > 0.0 && voxel.id != 170) {
            brightness = averageBrightness * (1.0 - (emissiveFactor / 255.0));
            b.emissive = true;
        }

        b.emission = emissionSpectra * brightness;
        if(b.emission > 0.0) {
            b.emissive = true;
        }

        int metalID = int(voxel.specular.g * 255.0 - 229.5);
        if(voxel.specular.g > 0.9999) {
            metalID = 20;
        }

        b.thickness = 0.0;
        b.thinFilm = false;
        b.filmIOR = complexFloat(1.333, 0.0);

        b.isLambertian = false;
        if(voxel.specular.g == 0.0 && b.roughness == 1.0) {
            b.isLambertian = true;
        }

        float ior = f0ToIOR(voxel.specular.g, wavelength);
        float k = 0.0;

        complexFloat iron = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 0), 0).rg);
        complexFloat gold = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 1), 0).rg);
        complexFloat aluminum = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 2), 0).rg);
        complexFloat chromium = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 3), 0).rg);
        complexFloat copper = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 4), 0).rg);
        complexFloat lead = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 5), 0).rg);
        complexFloat platinum = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 6), 0).rg);
        complexFloat silver = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 7), 0).rg);
        complexFloat bismuth = vectorToComplex(texelFetch(colortex14, ivec2(wavelength - 390.0, 9), 0).rg);

        if(b.isMetal) {
            //b.layerWeight = pow(saturate(voxel.thickness * 4.0 - 3.85), 18.0);
            b.layerWeight = 1.0;

            //ior = texelFetch(colortex14, ivec2(wavelength - 390.0, metalID), 0).r;
            //k = texelFetch(colortex14, ivec2(wavelength - 390.0, metalID), 0).g;

            //if(metalID == 20) {
                ior = GetN(b.metalAlbedo, saturate(b.metalAlbedo * 0.6 + 0.4), wavelength);
                k = GetK(b.metalAlbedo, ior);
            //}

            #ifdef ALBEDO_TINTED_METALS
                if(metalID != 20) {
                    float f0 = GetR(ior, k, wavelength) * b.metalAlbedo;
                    float g = GetG(ior, k, wavelength);
                    ior = GetN(f0, g, wavelength);
                    k = GetK(f0, ior);
                }
            #endif
        }

        b.n1 = complexFloat(Air(wavelength * 1e-3), 0.0);
        b.n2 = complexFloat(ior, k);

        if(b.thickness == 0.0) {
            //Film thickness is zero, there is no film.
            b.thinFilm = false;
        }

        return b;
    }
#endif