#if !defined LIB_FILMIC_PIECEWISE_TONEMAP
#define LIB_FILMIC_PIECEWISE_TONEMAP
    #define FixWhitepoint 0 //[0 1]
    #define ToeStrength 0.0 //[0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0]
    #define ToeLength 0.5 //[0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0]
    #define ShoulderStrength 0.0 //[0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0 1.1 1.2 1.3 1.4 1.5 1.6 1.7 1.8 1.9 2.0 2.1 2.2 2.3 2.4 2.5 2.6 2.7 2.8 2.9 3.0]
    #define ShoulderLength 0.5 //[0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0]
    #define ShoulderAngle 0.0 //[0.0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1.0]
    #define Gamma 1.0

    struct CurveParamsDirect
    {
        vec2 p0, p1;
        float w;
        vec2 overshoot;
        float gamma;
    };

    CurveParamsDirect CurveParamsDirect_new()
    {
        CurveParamsDirect self;
        self.p0 = vec2(0.25);
        self.p1 = vec2(0.75);
        self.w = 1.0;
        self.gamma = 1.0;
        self.overshoot = vec2(0.0);
        return self;
    }

    CurveParamsDirect CurveParamsDirect_new(
        vec2 p0, vec2 p1, float w, vec2 overshoot, float gamma)
    {
        CurveParamsDirect self;
        self.p0 = p0;
        self.p1 = p1;
        self.w = w;
        self.gamma = gamma;
        self.overshoot = overshoot;
        return self;
    }
    //#endregion

    //#region CurveSegment
    struct CurveSegment
    {
        vec2 offset, scale;
        float ln_a, b;
    };

    CurveSegment CurveSegment_new()
    {
        CurveSegment self;
        self.offset = vec2(0.0);
        self.scale = vec2(1.0);
        self.ln_a = 0.0;
        self.b = 1.0;
        return self;
    }

    CurveSegment CurveSegment_new(vec2 offset, vec2 scale, float ln_a, float b)
    {
        CurveSegment self;
        self.offset = offset;
        self.scale = scale;
        self.ln_a = ln_a;
        self.b = b;
        return self;
    }

    float CurveSegment_eval(CurveSegment self, float x)
    {
        vec2 p0 = vec2((x - self.offset.x) * self.scale.x, 0.0);

        if (p0.x > 0.0)
            p0.y = exp(self.ln_a + self.b * log(p0.x));
        
        return p0.y * self.scale.y + self.offset.y;
    }

    float CurveSegment_eval_inv(CurveSegment self, float y)
    {
        vec2 p0 = vec2(0.0, (y - self.offset.y) / self.scale.y);

        if (p0.y > 0.0)
            p0.x = exp((log(p0.y) - self.ln_a) / self.b);
        
        return p0.x / self.scale.x + self.offset.x;
    }
    //#endregion

    //#region FullCurve
    struct FullCurve
    {
        float w, inv_w;
        vec2 p0, p1;

        // CurveSegment[3]
        vec2 offset[3], scale[3];
        float ln_a[3], b[3];
    };

    CurveSegment FullCurve_get_segment(FullCurve self, int i)
    {
        return CurveSegment_new(
            self.offset[i], self.scale[i], self.ln_a[i], self.b[i]);
    }

    void FullCurve_set_segment(inout FullCurve self, int i, CurveSegment seg)
    {
        self.offset[i] = seg.offset;
        self.scale[i] = seg.scale;
        self.ln_a[i] = seg.ln_a;
        self.b[i] = seg.b;
    }

    FullCurve FullCurve_new()
    {
        FullCurve self;
        self.w = 1.0;
        self.inv_w = 1.0;
        self.p0 = vec2(0.25);
        self.p1 = vec2(0.75);

        for (int i = 0; i < 3; ++i)
            FullCurve_set_segment(self, i, CurveSegment_new());

        return self;
    }

    float FullCurve_eval(FullCurve self, float x)
    {
        float norm_x = x * self.inv_w;
        int index = (norm_x < self.p0.x) ? 0 : ((norm_x < self.p1.x) ? 1 : 2);
        CurveSegment segment = FullCurve_get_segment(self, index);
        return CurveSegment_eval(segment, norm_x);
    }

    float FullCurve_eval_inv(FullCurve self, float y)
    {
        int index = (y < self.p0.y) ? 0 : ((y < self.p1.y) ? 1 : 2);
        CurveSegment segment = FullCurve_get_segment(self, index);
        
        float norm_x = CurveSegment_eval_inv(segment, y);
        return norm_x * self.w;
    }
    //#endregion

    //#endregion

    //#region Functions

    // Returns vec2(ln_a, b)
    vec2 solve_a_b(vec2 p0, float m)
    {
        float b = (m * p0.x) / p0.y;
        float ln_a = log(p0.y) - b * log(p0.x);
        return vec2(ln_a, b);
    }

    // Returns vec2(m, b)
    vec2 as_slope_intercept(vec2 p0, vec2 p1)
    {
        vec2 d = p1 - p0;
        float m = (d.x == 0.0) ? 1.0 : d.y / d.x;
        float b = p0.y - p0.x * m;
        return vec2(m, b);
    }

    float eval_derivative_linear_gamma(float m, float b, float g, float x)
    {
        return g * m * pow(abs(m * x + b), g - 1.0);
    }

    FullCurve create_curve(CurveParamsDirect params)
    {
        FullCurve curve = FullCurve_new();
        curve.w = params.w;
        curve.inv_w = rcp(params.w);

        //params.w = 1.0;
        params.p0.x /= params.w;
        params.p1.x /= params.w;
        params.overshoot.x /= params.w;

        float toe_m = 0.0;
        float shoulder_m = 0.0;
        float endpoint_m = 0.0;
        {
            vec2 m_b = as_slope_intercept(params.p0, params.p1);
            float g = params.gamma;

            CurveSegment mid_segment = CurveSegment_new(
                vec2(-(m_b.y / m_b.x), 0.0), vec2(1.0), g * log(m_b.x), g);
            
            FullCurve_set_segment(curve, 1, mid_segment);

            toe_m = eval_derivative_linear_gamma(m_b.x, m_b.y, g, params.p0.x);
            shoulder_m = eval_derivative_linear_gamma(m_b.x, m_b.y, g, params.p1.x);

            params.p0.y = max(1e-5, pow(abs(params.p0.y), params.gamma));
            params.p1.y = max(1e-5, pow(abs(params.p1.y), params.gamma));

            params.overshoot.y = pow(abs(1.0 + params.overshoot.y), params.gamma) - 1.0;
        }

        curve.p0 = params.p0;
        curve.p1 = params.p1;

        {
            CurveSegment toe_segment = CurveSegment_new();
            toe_segment.offset = vec2(0.0);
            toe_segment.scale = vec2(1.0);

            vec2 a_b = solve_a_b(params.p0, toe_m);
            toe_segment.ln_a = a_b.x;
            toe_segment.b = a_b.y;
            FullCurve_set_segment(curve, 0, toe_segment);
        }

        {
            CurveSegment shoulder_segment = CurveSegment_new();

            vec2 p0 = (1.0 + params.overshoot) - params.p1;

            vec2 a_b = solve_a_b(p0, shoulder_m);

            shoulder_segment.offset = 1.0 + params.overshoot;

            shoulder_segment.scale = -vec2(1.0);
            shoulder_segment.ln_a = a_b.x;
            shoulder_segment.b = a_b.y;

            FullCurve_set_segment(curve, 2, shoulder_segment);
        }

        {
            float scale = CurveSegment_eval(FullCurve_get_segment(curve, 2), 1.0);
            float inv_scale = rcp(scale);

            curve.offset[0].y *= inv_scale;
            curve.scale[0].y *= inv_scale;

            curve.offset[1].y *= inv_scale;
            curve.scale[1].y *= inv_scale;

            curve.offset[2].y *= inv_scale;
            curve.scale[2].y *= inv_scale;
        }

        return curve;
    }

    CurveParamsDirect calc_direct_params_from_user()
    {
        CurveParamsDirect params = CurveParamsDirect_new();

        float perceptual_gamma = 2.2;
        float toe_length = pow(abs(ToeLength), perceptual_gamma);

        params.p0.x = toe_length * 0.5;
        params.p0.y = (1.0 - ToeStrength) * params.p0.x;

        float remaining_y = 1.0 - params.p0.y;

        float initial_w  = params.p0.x + remaining_y;

        float y1_offset = (1.0 - ShoulderLength) * remaining_y;
        params.p1 = params.p0 + y1_offset;

        float extra_w = exp2(ShoulderStrength) - 1.0;

        params.w = initial_w + extra_w;

        params.gamma = Gamma;

        params.overshoot = vec2(
            (params.w * 2.0) * ShoulderAngle * ShoulderStrength,
            0.5 * ShoulderAngle * ShoulderStrength);

        return params;
    }

    vec3 apply_filmic_curve(vec3 color) {
        CurveParamsDirect params = calc_direct_params_from_user();

        FullCurve curve = create_curve(params);

        color = vec3(
            FullCurve_eval(curve, color.r),
            FullCurve_eval(curve, color.g),
            FullCurve_eval(curve, color.b));

        switch (FixWhitepoint)
        {
            case 1:
                color /= FullCurve_eval(curve, 1.0);
                break;
            case 2:
                color *= FullCurve_eval_inv(curve, 1.0);
                break;
        }

        return color;
    }
#endif