#if !defined VOXEL_INTERSECTIONS_LIB
    #define VOXEL_INTERSECTIONS_LIB
        bool IntersectAABox(in vec3 rayPosition, in vec3 invRayDir, in vec3 minBounds, in vec3 maxBounds, in vec4 voxel0, in vec4 voxel1, out float dist, inout VoxelData voxel, in bool rotateUv) {
            // This will not intersect backfaces, or in other words this will not intersect the box from inside it.
            // You should use the version that rakes in radii instead of min & max bounds if you can, as it will be faster.
            vec3 minBoundsDist = (minBounds - rayPosition) * invRayDir;
            vec3 maxBoundsDist = (maxBounds - rayPosition) * invRayDir;

            vec3 minDists = min(minBoundsDist, maxBoundsDist);
            vec3 maxDists = max(minBoundsDist, maxBoundsDist);

            float maxofMin = maxof(minDists); // distance to frontfaces
            float minofMax = minof(maxDists); // distance to backfaces

            bool intersected = maxofMin > 0.0 && maxofMin <= minofMax;

            dist = maxofMin; // Distance to intersection (if the box was intersected)
            rayPosition = rayPosition + dist * (1.0/invRayDir);

            // Deterimine normal
            if (dist == minDists.x) {
                voxel.flatNormal = vec3((invRayDir.x < 0.0) ? 1.0 : -1.0, 0.0, 0.0);
                voxel.tbnMatrix = mat3(
                    vec3(0.0, 0.0, 1.0),
                    vec3(0.0, 1.0, 0.0),
                    voxel.flatNormal
                );
            } else if (dist == minDists.y) {
                voxel.flatNormal = vec3(0.0, (invRayDir.y < 0.0) ? 1.0 : -1.0, 0.0);
                voxel.tbnMatrix = mat3(
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 0.0, 1.0),
                    voxel.flatNormal
                );
            } else {
                voxel.flatNormal = vec3(0.0, 0.0, (invRayDir.z < 0.0) ? 1.0 : -1.0);
                voxel.tbnMatrix = mat3(
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 1.0, 0.0),
                    voxel.flatNormal
                );
            }

            voxel.id = int(voxel0.a * 255.0 + 0.5);

            voxel.uv  = vec2(dot(rayPosition, voxel.tbnMatrix[0]), dot(rayPosition, voxel.tbnMatrix[1]));
            voxel.uv  = fract(voxel.uv);
            voxel.uv += 256.0 * 15.0 * UnpackUnorm2x4(voxel1.y) + 255.0 * voxel1.zw;
            voxel.uv *= exp2(voxel1.x * 15.0) / textureSize(colortex6, 0);
            voxel.color = SrgbToLinear(textureLod(colortex6, voxel.uv, 0).rgb * voxel0.rgb);

            vec2 rawNormal = textureLod(colortex7, voxel.uv, 0).xy * 2.0 - 1.0;
            #ifndef MC_NORMAL_MAP
                voxel.normalMap = voxel.flatNormal;
                voxel.tangentNormal = normalize(voxel.normalMap * voxel.tbnMatrix);
            #else
                voxel.normalMap = normalize(voxel.tbnMatrix * vec3(rawNormal.xy, sqrt(saturate(1.0 - dot(rawNormal.xy, rawNormal.xy)))));
                voxel.tangentNormal = normalize(voxel.normalMap * voxel.tbnMatrix);
            #endif

            voxel.specular = textureLod(colortex8, voxel.uv, 0);

            voxel.ambientOcclusion = textureLod(colortex7, voxel.uv, 0).b * 2.0 - 1.0;

            voxel.opacity = textureLod(colortex6, voxel.uv, 0).a;

            bool alphaTest = voxel.opacity > max(RandNextF(), 26.5 / 255.0);

            if(voxel.id == 17) {
                alphaTest = true;
            }

            // Check for intersection & return whether or not we intersected
            return intersected && alphaTest;
        }

        bool IntersectAABox(in vec3 rayPosition, in vec3 invRayDir, in vec3 minBounds, in vec3 maxBounds, in vec4 voxel0, in vec4 voxel1, out float dist, inout VoxelData voxel) {
            return IntersectAABox(rayPosition, invRayDir, minBounds, maxBounds, voxel0, voxel1, dist, voxel, false);
        }

        bool IntersectAABoxBackFace(in vec3 rayPosition, in vec3 invRayDir, in vec3 minBounds, in vec3 maxBounds, in vec4 voxel0, in vec4 voxel1, out float dist, inout VoxelData voxel, in bool allowInside) {
        //The input "glass" only exists to avoid any issues.
        vec3 minBoundsDist = (minBounds - rayPosition) * invRayDir;
        vec3 maxBoundsDist = (maxBounds - rayPosition) * invRayDir;

        vec3 minDists = min(minBoundsDist, maxBoundsDist);
        vec3 maxDists = max(minBoundsDist, maxBoundsDist);

        float maxOfMin = maxof(minDists);
        float minOfMax = minof(maxDists);
        bool inside = maxOfMin < 0.001 && minOfMax > -0.001;

        dist = inside ? minOfMax : maxOfMin; // Distance to intersection (if the box was intersected)

        // Deterimine normal
        if(allowInside) {
            if (dist == (inside ? maxDists.x : minDists.x)) {
                voxel.flatNormal = vec3((invRayDir.x < 0.0 != inside) ? 1.0 : -1.0, 0.0, 0.0);
                voxel.tbnMatrix = mat3(
                    vec3(0.0, 0.0, 1.0),
                    vec3(0.0, 1.0, 0.0),
                    voxel.flatNormal
                );
            } else if (dist == (inside ? maxDists.y : minDists.y)) {
                voxel.flatNormal = vec3(0.0, (invRayDir.y < 0.0 != inside) ? 1.0 : -1.0, 0.0);
                voxel.tbnMatrix = mat3(
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 0.0, 1.0),
                    voxel.flatNormal
                );
            } else {
                voxel.flatNormal = vec3(0.0, 0.0, (invRayDir.z < 0.0 != inside) ? 1.0 : -1.0);
                voxel.tbnMatrix = mat3(
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 1.0, 0.0),
                    voxel.flatNormal
                );
            }
        } else {
            if (dist == (inside ? maxDists.x : minDists.x)) {
                voxel.flatNormal = vec3((invRayDir.x < 0.0) ? 1.0 : -1.0, 0.0, 0.0);
                voxel.tbnMatrix = mat3(
                    vec3(0.0, 0.0, 1.0),
                    vec3(0.0, 1.0, 0.0),
                    voxel.flatNormal
                );
            } else if (dist == (inside ? maxDists.y : minDists.y)) {
                voxel.flatNormal = vec3(0.0, (invRayDir.y < 0.0) ? 1.0 : -1.0, 0.0);
                voxel.tbnMatrix = mat3(
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 0.0, 1.0),
                    voxel.flatNormal
                );
            } else {
                voxel.flatNormal = vec3(0.0, 0.0, (invRayDir.z < 0.0) ? 1.0 : -1.0);
                voxel.tbnMatrix = mat3(
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 1.0, 0.0),
                    voxel.flatNormal
                );
            }
        }

        rayPosition = rayPosition + dist * (1.0/invRayDir);

        voxel.id = int(voxel0.a * 255.0 + 0.5);

        voxel.uv  = vec2(dot(rayPosition, voxel.tbnMatrix[0]), dot(rayPosition, voxel.tbnMatrix[1]));
        voxel.uv  = fract(voxel.uv);
        voxel.uv += 256.0 * 15.0 * UnpackUnorm2x4(voxel1.y) + 255.0 * voxel1.zw;
        voxel.uv *= exp2(voxel1.x * 15.0) / textureSize(colortex6, 0);
        voxel.color = SrgbToLinear(textureLod(colortex6, voxel.uv, 0).rgb * voxel0.rgb);

        vec2 rawNormal = textureLod(colortex7, voxel.uv, 0).xy * 2.0 - 1.0;
        #ifndef MC_NORMAL_MAP
            voxel.normalMap = voxel.flatNormal;
            voxel.tangentNormal = normalize(voxel.normalMap * voxel.tbnMatrix);
        #else
            voxel.normalMap = normalize(voxel.tbnMatrix * vec3(rawNormal.xy, sqrt(saturate(1.0 - dot(rawNormal.xy, rawNormal.xy)))));
            voxel.tangentNormal = normalize(voxel.normalMap * voxel.tbnMatrix);
        #endif

        voxel.specular = textureLod(colortex8, voxel.uv, 0);

        voxel.ambientOcclusion = textureLod(colortex7, voxel.uv, 0).b * 2.0 - 1.0;

        voxel.opacity = textureLod(colortex6, voxel.uv, 0).a;

        if(voxel.id == 156 || voxel.id == 157 || voxel.id == 158 || voxel.id == 159 || voxel.id == 161 || voxel.id == 166) {
            voxel.refractive = voxel.specular.g > 0.9 ? false : true;
        }

        // Determine & return whether or not we actually intersected the box
        return dist > 0.0 && maxOfMin <= minOfMax;
    }

    bool IntersectVoxel(vec3 relativeOrigin, vec3 direction, int id, in vec4 voxel0, in vec4 voxel1, inout float hitDist, inout VoxelData voxel) {
        vec3 inverseDirection = 1.0 / direction;

        if (id <= 16) {
            vec3 minBounds, maxBounds;
            minBounds = vec3(0);
            maxBounds = vec3(1, id * 0.0625 - 0.0625, 1);

            return IntersectAABox(relativeOrigin, inverseDirection, minBounds, maxBounds, voxel0, voxel1, hitDist, voxel);
        } else if (id <= 96) {
            vec3[97 - 17][2] bounds1 = vec3[97 - 17][2](
                // Fluids
                // This is currently just so they don't look completely awful
                vec3[2](vec3(0), vec3(1, 8.0/9.0, 1)),
                vec3[2](vec3(0), vec3(1, 8.0/9.0, 1)),

                // Slabs (Top)
                vec3[2](vec3(0, 0.5, 0), vec3(1)),
                // Torches (not on walls)
                vec3[2](vec3(0.4375, 0.0, 0.4375), vec3(0.5625, 0.625, 0.5625)),

                // Pressure Plates (not pressed)
                vec3[2](vec3(0.0625, 0, 0.0625), vec3(0.9375, 0.0625, 0.9375)),
                // Pressure Plates (pressed)
                vec3[2](vec3(0.0625, 0, 0.0625), vec3(0.9375, 0.03125, 0.9375)),

                // Doors & Trapdoors
                vec3[2](vec3(0), vec3(0.1875, 1, 1)),
                vec3[2](vec3(0.8125, 0, 0), vec3(1)),
                vec3[2](vec3(0), vec3(1, 1, 0.1875)),
                vec3[2](vec3(0, 0, 0.8125), vec3(1)),
                vec3[2](vec3(0, 0.8125, 0), vec3(1)),

                // Buttons
                vec3[2](vec3(0.0000, 0.375, 0.3125), vec3(0.1250, 0.625, 0.6875)),
                vec3[2](vec3(0.8750, 0.375, 0.3125), vec3(1.0000, 0.625, 0.6875)),
                vec3[2](vec3(0.3125, 0.375, 0.0000), vec3(0.6875, 0.625, 0.1250)),
                vec3[2](vec3(0.3250, 0.375, 0.8750), vec3(0.6875, 0.625, 1.0000)),
                vec3[2](vec3(0.3750, 0.000, 0.3125), vec3(0.6250, 0.125, 0.6875)),
                vec3[2](vec3(0.3125, 0.000, 0.3750), vec3(0.6875, 0.125, 0.6250)),
                vec3[2](vec3(0.3750, 0.875, 0.3125), vec3(0.6250, 1.000, 0.6875)),
                vec3[2](vec3(0.3125, 0.875, 0.3750), vec3(0.6875, 1.000, 0.6250)),

                vec3[2](vec3(0.0000, 0.3750, 0.3125), vec3(0.0625, 0.6250, 0.6875)),
                vec3[2](vec3(0.9375, 0.3750, 0.3125), vec3(1.0000, 0.6250, 0.6875)),
                vec3[2](vec3(0.3125, 0.3750, 0.0000), vec3(0.6875, 0.6250, 0.0625)),
                vec3[2](vec3(0.3125, 0.3750, 0.9375), vec3(0.6875, 0.6250, 1.0000)),
                vec3[2](vec3(0.3750, 0.0000, 0.3125), vec3(0.6250, 0.0625, 0.6875)),
                vec3[2](vec3(0.3125, 0.0000, 0.3750), vec3(0.6875, 0.0625, 0.6250)),
                vec3[2](vec3(0.3750, 0.9375, 0.3125), vec3(0.6250, 1.0000, 0.6875)),
                vec3[2](vec3(0.3125, 0.9375, 0.3750), vec3(0.6875, 1.0000, 0.6250)),

                // Cactus
                vec3[2](vec3(0.0625, 0, 0.0625), vec3(0.9375, 1, 0.9375)),

                // Nether portal
                vec3[2](vec3(0, 0, 0.375), vec3(1, 1, 0.625)),
                vec3[2](vec3(0.375, 0, 0), vec3(0.625, 1, 1)),

                // Ladders, some states of vines
                vec3[2](vec3(0.05, 0, 0), vec3(0.05, 1, 1)), // east
                vec3[2](vec3(0.95, 0, 0), vec3(0.95, 1, 1)), // west
                vec3[2](vec3(0, 0, 0.05), vec3(1, 1, 0.05)), // south
                vec3[2](vec3(0, 0, 0.95), vec3(1, 1, 0.95)), // north

                // Glass panes, first box (also iron bars until i get around to giving them their own model)
                vec3[2](vec3(0.4375, 0, 0.4375), vec3(0.5625, 1, 0.5625)), // none
                vec3[2](vec3(0.4375, 0, 0     ), vec3(0.5625, 1, 0.5   )), // north
                vec3[2](vec3(0.4375, 0, 0.5   ), vec3(0.5625, 1, 1     )), // south
                vec3[2](vec3(0,      0, 0.4375), vec3(0.5,    1, 0.5625)), // west
                vec3[2](vec3(0.5,    0, 0.4375), vec3(1,      1, 0.5625)), // east
                vec3[2](vec3(0.4375, 0, 0     ), vec3(0.5625, 1, 1     )), // north-south
                vec3[2](vec3(0,      0, 0.4375), vec3(1,      1, 0.5625)), // east-west

                // End rods
                vec3[2](vec3(0.0625, 0.4375, 0.4375), vec3(1,      0.5625, 0.5625)),
                vec3[2](vec3(0,      0.4375, 0.4375), vec3(0.9375, 0.5625, 0.5625)),
                vec3[2](vec3(0.4375, 0.0625, 0.4375), vec3(0.5625, 1,      0.5625)),
                vec3[2](vec3(0.4375, 0,      0.4375), vec3(0.5625, 0.9375, 0.5625)),
                vec3[2](vec3(0.4375, 0.4375, 0.0625), vec3(0.5625, 0.5625, 1     )),
                vec3[2](vec3(0.4375, 0.4375, 0     ), vec3(0.5625, 0.5625, 0.9375)),

                // Stairs (straight)
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),

                // Stairs (outer corner)
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),

                // Glass panes with second box (also iron bars until i get around to giving them their own model)
                vec3[2](vec3(0,      0, 0.4375), vec3(0.5625, 1, 0.5625)), // north-west
                vec3[2](vec3(0,      0, 0.4375), vec3(0.5625, 1, 0.5625)), // south-west
                vec3[2](vec3(0.4375, 0, 0.4375), vec3(1,      1, 0.5625)), // north-east
                vec3[2](vec3(0.4375, 0, 0.4375), vec3(1,      1, 0.5625)), // south-east

                vec3[2](vec3(0,      0, 0.4375), vec3(0.5625, 1, 0.5625)), // north, south, west
                vec3[2](vec3(0.4375, 0, 0.4375), vec3(1,      1, 0.5625)), // north, south, east
                vec3[2](vec3(0,      0, 0.4375), vec3(1,      1, 0.5625)), // north, east, west
                vec3[2](vec3(0,      0, 0.4375), vec3(1,      1, 0.5625)), // south, east, west

                vec3[2](vec3(0,      0, 0.4375), vec3(1,      1, 0.5625)), // cross

                // Stairs (inner corner)
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     )),
                vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      1     ))
            );

            bool intersected = IntersectAABox(relativeOrigin, inverseDirection, bounds1[id - 17][0], bounds1[id - 17][1], voxel0, voxel1, hitDist, voxel);

            if (58 <= id) {
                vec3[97 - 58][2] bounds2 = vec3[97 - 58][2](
                    // End rods
                    vec3[2](vec3(0,      0.375,  0.375 ), vec3(0.0625, 0.625,  0.625 )),
                    vec3[2](vec3(0.9375, 0.375,  0.375 ), vec3(1,      0.625,  0.625 )),
                    vec3[2](vec3(0.375,  0,      0.375 ), vec3(0.625,  0.0625, 0.625 )),
                    vec3[2](vec3(0.375,  0.9375, 0.375 ), vec3(0.625,  1,      0.625 )),
                    vec3[2](vec3(0.375,  0.375,  0     ), vec3(0.625,  0.625,  0.0625)),
                    vec3[2](vec3(0.375,  0.375,  0.9375), vec3(0.625,  0.625,  1     )),

                    // Stairs (straight)
                    vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      0.5   )),
                    vec3[2](vec3(0,      0.5,    0.5   ), vec3(1,      1,      1     )),
                    vec3[2](vec3(0.5,    0.5,    0     ), vec3(1,      1,      1     )),
                    vec3[2](vec3(0,      0.5,    0     ), vec3(0.5,    1,      1     )),
                    vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    0.5   )),
                    vec3[2](vec3(0,      0,      0.5   ), vec3(1,      0.5,    1     )),
                    vec3[2](vec3(0.5,    0,      0     ), vec3(1,      0.5,    1     )),
                    vec3[2](vec3(0,      0,      0     ), vec3(0.5,    0.5,    1     )),

                    // Stairs (outer corner)
                    vec3[2](vec3(0.5,    0.5,    0     ), vec3(1,      1,      0.5   )),
                    vec3[2](vec3(0.5,    0.5,    0.5   ), vec3(1,      1,      1     )),
                    vec3[2](vec3(0,      0.5,    0     ), vec3(0.5,    1,      0.5   )),
                    vec3[2](vec3(0,      0.5,    0.5   ), vec3(0.5,    1,      1     )),
                    vec3[2](vec3(0.5,    0,      0     ), vec3(1,      0.5,    0.5   )),
                    vec3[2](vec3(0.5,    0,      0.5   ), vec3(1,      0.5,    1     )),
                    vec3[2](vec3(0,      0,      0     ), vec3(0.5,    0.5,    0.5   )),
                    vec3[2](vec3(0,      0,      0.5   ), vec3(0.5,    0.5,    1     )),

                    // Glass panes with second box (also iron bars until i get around to giving them their own model)
                    vec3[2](vec3(0.4375, 0, 0     ), vec3(0.5625, 1, 0.5625)), // north-west
                    vec3[2](vec3(0.4375, 0, 0.4375), vec3(0.5625, 1, 1     )), // south-west
                    vec3[2](vec3(0.4375, 0, 0     ), vec3(0.5625, 1, 0.5625)), // north-east
                    vec3[2](vec3(0.4375, 0, 0.4375), vec3(0.5625, 1, 1     )), // south-east

                    vec3[2](vec3(0.4375, 0, 0     ), vec3(0.5625, 1, 1     )), // north, south, west
                    vec3[2](vec3(0.4375, 0, 0     ), vec3(0.5625, 1, 1     )), // north, south, east
                    vec3[2](vec3(0.4375, 0, 0     ), vec3(0.5625, 1, 0.5625)), // north, east, west
                    vec3[2](vec3(0.4375, 0, 0.4375), vec3(0.5625, 1, 1     )), // south, east, west

                    vec3[2](vec3(0.4375, 0, 0     ), vec3(0.5625, 1, 1     )), // cross

                    // Stairs (inner corner)
                    vec3[2](vec3(0,      0.5,    0     ), vec3(1,      1,      0.5   )),
                    vec3[2](vec3(0,      0.5,    0.5   ), vec3(1,      1,      1     )),
                    vec3[2](vec3(0,      0.5,    0     ), vec3(0.5,    1,      1     )),
                    vec3[2](vec3(0,      0.5,    0     ), vec3(0.5,    1,      1     )),
                    vec3[2](vec3(0,      0,      0     ), vec3(1,      0.5,    0.5   )),
                    vec3[2](vec3(0,      0,      0.5   ), vec3(1,      0.5,    1     )),
                    vec3[2](vec3(0,      0,      0     ), vec3(0.5,    0.5,    1     )),
                    vec3[2](vec3(0,      0,      0     ), vec3(0.5,    0.5,    1     ))
                );

                float tDist;
                VoxelData tempVoxel;
                if (IntersectAABox(relativeOrigin, inverseDirection, bounds2[id - 58][0], bounds2[id - 58][1], voxel0, voxel1, tDist, tempVoxel)) {
                    if (!intersected || tDist < hitDist) {
                        hitDist = tDist;
                        voxel = tempVoxel;
                    }
                    intersected = true;
                }
            }

            if (89 <= id) {
                vec3[97 - 89][2] bounds3 = vec3[97 - 89][2](
                    // Stairs (inner corner)
                    vec3[2](vec3(0.5,    0.5,    0.5   ), vec3(1,      1,      1     )),
                    vec3[2](vec3(0.5,    0.5,    0     ), vec3(1,      1,      0.5   )),
                    vec3[2](vec3(0.5,    0.5,    0     ), vec3(1,      1,      0.5   )),
                    vec3[2](vec3(0.5,    0.5,    0.5   ), vec3(1,      1,      1     )),
                    vec3[2](vec3(0.5,    0,      0.5   ), vec3(1,      0.5,    1     )),
                    vec3[2](vec3(0.5,    0,      0     ), vec3(1,      0.5,    0.5   )),
                    vec3[2](vec3(0.5,    0,      0     ), vec3(1,      0.5,    0.5   )),
                    vec3[2](vec3(0.5,    0,      0.5   ), vec3(1,      0.5,    1     ))
                );

                float tDist;
                VoxelData tempVoxel;
                if (IntersectAABox(relativeOrigin, inverseDirection, bounds3[id - 89][0], bounds3[id - 89][1], voxel0, voxel1, tDist, tempVoxel)) {
                    if (!intersected || tDist < hitDist) {
                        hitDist = tDist;
                        voxel = tempVoxel;
                    }
                    intersected = true;
                }
            }

            return intersected;
        }
    }

    //*
    bool IntersectCrossBlock(vec3 relativeOrigin, vec3 direction, in vec4 voxel0, in vec4 voxel1, out float hitDist, inout VoxelData voxel) {
        vec3 inverseDirection = 1.0 / direction;

        vec3 rotationOrigin = vec3(0.5);
        vec3 rotationVector = vec3(0.0, 1.0, 0.0);

        mat3 m0 = GetRotationMatrix(rotationVector, radians(45.0));

        hitDist = 1.0 / 0.0;

        float dist0;
        bool hit0 = IntersectPlate(m0 * (relativeOrigin - rotationOrigin), m0 * direction, vec2(0.5), dist0);
        if(hit0 && dist0 < hitDist) {
            vec3 hitPosition = relativeOrigin + direction * dist0;
            vec2 tUv  = vec2(dot(hitPosition, transpose(m0)[0]) + 0.5, -dot(hitPosition, transpose(m0)[1]));
                 tUv  = fract(tUv);
                 tUv += 256.0 * 15.0 * UnpackUnorm2x4(voxel1.y) + 255.0 * voxel1.zw;
                 tUv *= exp2(voxel1.x * 15.0) / textureSize(colortex6, 0);

            if (textureLod(colortex6, tUv, 0).a > max(RandNextF(), 26.5 / 255.0)) {
                hitDist = dist0;
                voxel.tbnMatrix = transpose(m0);
                voxel.tbnMatrix[1] = -voxel.tbnMatrix[1];
                voxel.tbnMatrix[2] *= sign(-(m0 * direction).z);
                voxel.uv = tUv;
            } else {
                hit0 = false;
            }
        }

        mat3 m1 = GetRotationMatrix(rotationVector, radians(-45.0));

        float dist1;
        bool hit1 = IntersectPlate(m1 * (relativeOrigin - rotationOrigin), m1 * direction, vec2(0.5), dist1);
        if(hit1 && dist1 < hitDist) {
            vec3 hitPosition = relativeOrigin + direction * dist1;
            vec2 tUv  = vec2(dot(hitPosition, transpose(m1)[0]) - 0.238, -dot(hitPosition, transpose(m1)[1]));
                 tUv  = fract(tUv);
                 tUv += 256.0 * 15.0 * UnpackUnorm2x4(voxel1.y) + 255.0 * voxel1.zw;
                 tUv *= exp2(voxel1.x * 15.0) / textureSize(colortex6, 0);

            if (textureLod(colortex6, tUv, 0).a > max(RandNextF(), 26.5 / 255.0)) {
                hitDist = dist1;
                voxel.tbnMatrix = transpose(m1);
                voxel.tbnMatrix[1] = -voxel.tbnMatrix[1];
                voxel.tbnMatrix[2] *= sign(-(m1 * direction).z);
                voxel.uv = tUv;
            } else {
                hit1 = false;
            }
        }

        voxel.id = int(voxel0.a * 255.0 + 0.5);
        voxel.color = SrgbToLinear(textureLod(colortex6, voxel.uv, 0).rgb * voxel0.rgb);
        voxel.specular = textureLod(colortex8, voxel.uv, 0);
        voxel.ambientOcclusion = textureLod(colortex7, voxel.uv, 0).b * 2.0 - 1.0;

        voxel.opacity = textureLod(colortex6, voxel.uv, 0).a;

        voxel.flatNormal = voxel.tbnMatrix[2];
        vec2 rawNormal = textureLod(colortex7, voxel.uv, 0).xy * 2.0 - 1.0;
        #ifndef MC_NORMAL_MAP
            voxel.normalMap = voxel.flatNormal;
            voxel.tangentNormal = normalize(voxel.normalMap * voxel.tbnMatrix);
        #else
            voxel.normalMap = normalize(voxel.tbnMatrix * vec3(rawNormal.xy, sqrt(saturate(1.0 - dot(rawNormal.xy, rawNormal.xy)))));
            voxel.tangentNormal = normalize(voxel.normalMap * voxel.tbnMatrix);
        #endif

        return hit0 || hit1;
    }
    //*/

    bool IntersectSlime(vec3 relativeOrigin, vec3 inverseDirection, in vec4 voxel0, in vec4 voxel1, out float dist, inout VoxelData voxel) {
        vec3 minBounds, maxBounds;
        minBounds = vec3(0);
        maxBounds = vec3(1);

        bool hit = IntersectAABoxBackFace(relativeOrigin, inverseDirection, minBounds, maxBounds, voxel0, voxel1, dist, voxel, true);

        float tDist;
        VoxelData tVoxel;
        if(IntersectAABoxBackFace(relativeOrigin, inverseDirection, vec3(0.25), vec3(0.75), voxel0, voxel1, tDist, tVoxel, true)) {
            tVoxel.id = 999; //Inner Slime Cube
            if(!hit || tDist < dist) {
                dist = tDist;
                voxel = tVoxel;
            }
            hit = true;
        }

        return hit;
    }
#endif