#if !defined VOXEL_RAYTRACER_LIB
#define VOXEL_RAYTRACER_LIB
    #define RAYTRACE_MAX_T 200
    struct VoxelData {
        mat3 tbnMatrix;

        vec4 specular;

        vec3 color;
        vec3 flatNormal;
        vec3 tangentNormal;
        vec3 normalMap;

        vec2 uv;

        float opacity;
        float ambientOcclusion;
        float thickness;

        int id;

        bool refractive;
        bool sameMaterial;
    };

    #include "/lib/fragment/path/parallax.glsl"
    #include "intersections.glsl"

    vec2 DistortFBM2D(in vec2 position, int octaves, float amplitude, float gain) {
        vec2 value = vec2(texture(noisetex, position).r, texture(noisetex, position + vec2(3.33, 5.71)).r) * amplitude;
        float inv_weight = 1.0;
        for (int i = 1; i < octaves; i++){
            amplitude *= gain;
            position *= pi;
            value += vec2(texture(noisetex, position).r, texture(noisetex, position + vec2(3.33, 5.71)).r) * amplitude;
            inv_weight += amplitude;
        }
        return value / inv_weight;
    }

    float FBM(vec2 position) {
        float amplitude = 1.0;
        float V = amplitude * texture(noisetex, position).r; 

        float persistance = 0.47;
        float lacunarity = 3.25;
        
        float invWeight = 0.0;
        for (int i = 0; i < 6; ++i) {
            V += amplitude * texture(noisetex, position).r; 
            amplitude *= persistance; 
            position *= lacunarity;
            invWeight += amplitude;
        }
        
        return V / invWeight; 
    }

    bool RaytraceVoxel(vec3 position, ivec3 index, vec3 direction, out vec3 hitPosition, out ivec3 hitIndex, out float hitDistance, out VoxelData data) {
        //https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.42.3443&rep=rep1&type=pdf
        if(OutsideVoxelmap(index)) {
            return false;
        }

        vec4 voxel0 = unpackUnorm4x8(texelFetch(shadowcolor0, CalculateVoxelmapIdx(index), 0).r);
        vec4 voxel1 = texelFetch(shadowcolor1, CalculateVoxelmapIdx(index), 0);

        bool hit = false;
        int id = int(voxel0.a * 255.0 + 0.5);
        if(voxel0.a < 1.0) {
            if(id == 156 || id == 157 || id == 158 || id == 159 || id == 161) {
                if(IntersectAABoxBackFace(position - index, 1.0 / direction, vec3(0.0), vec3(1.0), voxel0, voxel1, hitDistance, data, true)) {
                    hitPosition = position + hitDistance * direction;
                    hitIndex    = index;

                    data.sameMaterial = false;

                    return true;
                }
            } else if(id == 166) {
                if(IntersectSlime(position - index, 1.0 / direction, voxel0, voxel1, hitDistance, data)) {
                    hitPosition = position + hitDistance * direction;
                    hitIndex    = index;

                    data.sameMaterial = false;

                    return true;
                }
            }
        }

        vec3 nextT;
        vec3 deltaDist = 1.0 / abs(direction);
        ivec3 deltaSign;
        for(int i = 0; i < 3; ++i) {
            if(direction[i] < 0.0) {
                deltaSign[i] = -1;
                nextT[i] = (position[i] - index[i]) * deltaDist[i];
            } else if(direction[i] > 0.0) {
                deltaSign[i] = 1;
                nextT[i] = (index[i] - position[i]) * deltaDist[i] + deltaDist[i];
            } else {
                deltaSign[i] = 0;
                nextT[i] = uintBitsToFloat(0x7f800000u);
            }
        }

	    if (deltaSign.x == 0 && deltaSign.y == 0 && deltaSign.z == 0) { return false; }

        data.sameMaterial = false;

        int previousID = 0;
        int currentID  = int(voxel0.a * 255.0 + 0.5);
        do {
            if(nextT.x < nextT.y && nextT.x < nextT.z) {
                hitDistance = nextT.x;
                data.flatNormal = vec3(-deltaSign.x, 0.0, 0.0);
                data.tbnMatrix = mat3(
                    vec3(0.0, 0.0, 1.0),
                    vec3(0.0, 1.0, 0.0),
                    data.flatNormal
                );

                data.sameMaterial = previousID == currentID;

                index.x += deltaSign.x;
                nextT.x += deltaDist.x;
            } else if(nextT.y < nextT.z) {
                hitDistance = nextT.y;
                data.flatNormal = vec3(0.0, -deltaSign.y, 0.0);
                data.tbnMatrix = mat3(
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 0.0, 1.0),
                    data.flatNormal
                );

                data.sameMaterial = previousID == currentID;

                index.y += deltaSign.y;
                nextT.y += deltaDist.y;
            } else {
                hitDistance = nextT.z;
                data.flatNormal = vec3(0.0, 0.0, -deltaSign.z);
                data.tbnMatrix = mat3(
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 1.0, 0.0),
                    data.flatNormal
                );

                data.sameMaterial = previousID == currentID;

                index.z += deltaSign.z;
                nextT.z += deltaDist.z;
            }

            if(hitDistance > RAYTRACE_MAX_T) {
                break;
            }
            if(OutsideVoxelmap(index)) {
                break;
            }

            voxel0 = unpackUnorm4x8(texelFetch(shadowcolor0, CalculateVoxelmapIdx(index), 0).r);
            voxel1 = texelFetch(shadowcolor1, CalculateVoxelmapIdx(index), 0);

            previousID = currentID;
            currentID  = int(voxel0.a * 255.0 + 0.5);

            if(voxel0.a < 1.0) {
                if(currentID == 156 || currentID == 157 || currentID == 158 || currentID == 159 || currentID == 161) {
                    IntersectAABoxBackFace(position - index, 1.0 / direction, vec3(0.0), vec3(1.0), voxel0, voxel1, hitDistance, data, true);

                    hitPosition = position + hitDistance * direction;
                    hitIndex = index;

                    data.sameMaterial = previousID == 156 || previousID == 157 || previousID == 158 || previousID == 159 || previousID == 161;

                    hit = true;
                } else if(currentID == 166) {
                    IntersectSlime(position - index, 1.0 / direction, voxel0, voxel1, hitDistance, data);

                    hitPosition = position + hitDistance * direction;
                    hitIndex = index;

                    data.sameMaterial = previousID == 166;

                    hit = true;
                } else {
                    hitPosition = position + hitDistance * direction;
                    hitIndex = index;

                    data.id = currentID;

                    vec2 tileIdx = 256.0 * 15.0 * UnpackUnorm2x4(voxel1.y) + 255.0 * voxel1.zw;
                    ivec2 tileSize = ivec2(exp2(voxel1.x * 15.0));
                    ivec2 tileOffset = ivec2(tileIdx * tileSize);

                    mat2 tileInfo = mat2(
                        exp2(voxel1.x * 15.0) / textureSize(colortex6, 0),
                        tileIdx * (exp2(voxel1.x * 15.0) / textureSize(colortex6, 0))
                    );

                    data.uv  = vec2(dot(hitPosition, data.tbnMatrix[0]), dot(hitPosition, data.tbnMatrix[1]));
                    data.uv  = fract(data.uv);
                    data.uv += tileIdx;
                    data.uv *= tileInfo[0];

                    #ifdef PARALLAX_MAPPING
                        vec3 parallaxHitPosition;
                        ivec2 hitTexel;
                        bool hitPOM = RaytraceParallax(direction * data.tbnMatrix, vec3((data.uv - tileInfo[1]) / tileInfo[0], 1.0), tileOffset, tileSize, ivec2(vec3((data.uv - tileInfo[1]) / tileInfo[0], 1.0).xy * tileSize), true, parallaxHitPosition, data.flatNormal, hitTexel);

                        vec3 offset    = parallaxHitPosition - vec3((data.uv - tileInfo[1]) / tileInfo[0], 1.0);
                             offset.z *= PARALLAX_DEPTH;

                        data.uv = vec2(hitTexel % tileSize + tileOffset) / vec2(textureSize(colortex6, 0));

                        data.flatNormal = data.tbnMatrix * data.flatNormal;
                        data.tbnMatrix  = GetRotationMatrix(vec3(0.0, 0.0, 1.0), data.flatNormal);

                        position += data.tbnMatrix * offset;

                        hitPosition = position + hitDistance * direction;
                    #else
                        bool hitPOM = true;
                    #endif

                    data.color = SrgbToLinear(textureLod(colortex6, data.uv, 0).rgb * voxel0.rgb);

                    vec2 rawNormal = textureLod(colortex7, data.uv, 0).xy * 2.0 - 1.0;
                    #ifndef MC_NORMAL_MAP
                        data.normalMap = normalize(data.tbnMatrix * vec3(0, 0, 1));
                        data.tangentNormal = normalize(data.normalMap * data.tbnMatrix);
                    #else
                        data.normalMap = normalize(data.tbnMatrix * vec3(rawNormal.xy, sqrt(saturate(1.0 - dot(rawNormal.xy, rawNormal.xy)))));
                        data.tangentNormal = normalize(data.normalMap * data.tbnMatrix);
                    #endif

                    data.specular = textureLod(colortex8, data.uv, 0);

                    data.ambientOcclusion = textureLod(colortex7, data.uv, 0).b;

                    data.opacity = textureLod(colortex6, data.uv, 0).a;

                    data.thickness = FBM((vec2(dot(SceneSpaceFromVoxelSpace(hitPosition) + cameraPosition, data.tbnMatrix[0]), dot(SceneSpaceFromVoxelSpace(hitPosition) + cameraPosition, data.tbnMatrix[1])) / noiseTextureResolution) * 2.0) - 0.2;

                    bool alphaTest = data.opacity > max(RandNextF(), 26.5 / 255.0);
                    if(data.id == 17) {
                        alphaTest = true;
                    }

                    data.sameMaterial = previousID == currentID;

                    hit = hitPOM && alphaTest;
                }
            }
        } while(!hit);

        return hit;
    }
#endif