#if !defined VOXEL_RAYTRACER_LIB
#define VOXEL_RAYTRACER_LIB
    #define RAYTRACE_MAX_T 100
    struct VoxelData {
        mat3 tbnMatrix;

        vec4 specular;

        vec3 color;
        vec3 flatNormal;
        vec3 tangentNormal;
        vec3 normalMap;

        vec2 uv;

        float opacity;
        float ambientOcclusion;
        float thickness;

        int id;

        bool refractive;
        bool sameMaterial;
    };

    float FBM(vec2 position) {
        float amplitude = 1.0;
        float V = amplitude * texture(noisetex, position).r; 

        float persistance = 0.47;
        float lacunarity = 3.25;

        float invWeight = 0.0;
        for (int i = 0; i < 6; ++i) {
            V += amplitude * texture(noisetex, position).r; 
            amplitude *= persistance; 
            position *= lacunarity;
            invWeight += amplitude;
        }
        
        return V / invWeight; 
    }

    float GerstnerWaves(vec2 coord, float time, float waveAmplitude, float waveLength, vec2 waveDirection) {
        const float g = 9.81;
        
        float k = (3.14159 * 2.0) / waveLength;
        float w = sqrt(g * k);
        float x = w * time - k * dot(waveDirection, coord);
        float wave = sin(x) * 0.5 + 0.5;
        float h = waveAmplitude * wave;
        return h;
    }

    vec2 GerstnerWavesDisp(vec2 coord, float time, float waveAmplitude, float waveLength, vec2 waveDirection) {
        const float g = 9.81;
        
        float k = (3.14159 * 2.0) / waveLength;
        float w = sqrt(g * k);
        float x = w * time - k * dot(waveDirection, coord);
        float wave = cos(x) * 0.5 + 0.5;
        float h = waveAmplitude * wave;
        return waveDirection * h;
    }

    vec3 CalculateGerstnerWavesHeight(in vec2 position) {
        vec2 waveDirection = vec2(0.2, 1.0);
        float waveLength = 4.0;
        float waveAmplitude = 0.02;
        float waveSteepness = 2.0;
        float a = 3.14159;
        float time = 1.0;
        vec2 positionNoise = (position * 5e-7);
        float height = 0.0;
        float dispX = 0.0;
        float dispZ = 0.0;
        for(int i = 0; i < 8; ++i) {
            a *= radians(32.0);

            vec2 noise = vec2(FBM(positionNoise)) * 0.5 + 0.5;

            height += -GerstnerWaves(position, time, waveAmplitude, waveLength, normalize(waveDirection));
            dispX += GerstnerWavesDisp(position, time, waveAmplitude, waveLength, normalize(waveDirection)).x * waveSteepness;
            dispZ += GerstnerWavesDisp(position, time, waveAmplitude, waveLength, normalize(waveDirection)).y * waveSteepness;

            position -= GerstnerWavesDisp(position, time, waveAmplitude, waveLength, normalize(waveDirection)) * waveSteepness;

            waveLength *= 0.95;
            waveAmplitude *= 0.90;
            waveSteepness *= 0.8;
            waveDirection *= Rotate(a * noise.x);
            positionNoise *= Rotate(a * noise.x);
            positionNoise *= 1.1;
        }
        return vec3(dispX, height, dispZ);
    }

    mat2 JacobianMatrixGerstner(in vec2 position) {
        vec2 px = vec2(position.x + 1, position.y);
        vec2 nx = vec2(position.x - 1, position.y);
        vec2 pz = vec2(position.x, position.y + 1);
        vec2 nz = vec2(position.x, position.y - 1);
        vec2 v_px = CalculateGerstnerWavesHeight(px).xz;
        vec2 v_nx = CalculateGerstnerWavesHeight(nx).xz;
        vec2 v_pz = CalculateGerstnerWavesHeight(pz).xz;
        vec2 v_nz = CalculateGerstnerWavesHeight(nz).xz;
        mat2 jacobianMatrix = mat2(
            v_px - v_nx,
            v_pz - v_nz
        ) + mat2(1.0);
        return jacobianMatrix;
    }

    vec3 WaterNormal(vec2 position) {
        const float dist = 4e-3;

        vec2 diffs;
        diffs.x = CalculateGerstnerWavesHeight(position + vec2( dist, -dist)).y;
        diffs.y = CalculateGerstnerWavesHeight(position + vec2(-dist, dist)).y;
        diffs  -= CalculateGerstnerWavesHeight(position + vec2(-dist, -dist)).y;
        diffs  *= transpose(inverse(JacobianMatrixGerstner(position)));

        vec3 wavenormal = vec3(-2.0 * dist, 4.0 * dist * dist, -2.0 * (dist * dist + dist));
        wavenormal.xz *= diffs;

        return wavenormal.xzy;
    }

    #include "/lib/fragment/path/parallax.glsl"
    #include "intersections.glsl"

    #define RELIABLE_VOXEL_TRAVERSAL // Slower, but practically has no leaks through corners or cases where it gets stuck

    uint traversalSteps = 0u;

    bool TraverseVoxelsNooctree(
        vec3  rayOrigin,    // Starting point of ray
        ivec3 rayOriginIdx, // Starting voxel
        vec3  rayDirection, // Traversal direction

        out vec3  hitPosition, // Intersection point of ray & voxel
        out ivec3 hitIdx,      // Intersected voxel
        out vec3  hitNormal    // Normal of voxel face that was intersected
    ) {
        // Validate rayDirection
        if (any(isnan(rayDirection)) || (rayDirection == vec3(0.0))) { return false; }

        // Fix inf directions
        // These probably aren't intended, but if they are they should be handled properly.
        bvec3 infComponent = isinf(rayDirection);
        if (any(infComponent)) {
            rayDirection = mix(vec3(0.0), SignExtract(rayDirection), infComponent);
        }

        // Init internal variables
        vec3  rayPosition = rayOrigin;
        ivec3 rayIdx      = rayOriginIdx;

        bvec3 negativeDir =    lessThan(rayDirection, vec3(0.0));
        bvec3 positiveDir = greaterThan(rayDirection, vec3(0.0));

        ivec3 stepv = ivec3(positiveDir) - ivec3(negativeDir);

        // Handle out-of-volume origin
        if (OutsideVoxelmap(rayOriginIdx)) {
            // Is there a more robust way of doing this?
            // Current method feels like it might be a bit unreliable in near-edge cases...

            // Set up "near" and "far" planes of voxel volume
            // Cast to uvec3 because of broken nvidia driver, will be removed
            uvec3 nearPlanes = uvec3(mix(minBounds.xyx, maxBounds.xyx, negativeDir));
            uvec3 farPlanes  = uvec3(mix(minBounds.xyx, maxBounds.xyx, positiveDir));

            vec3 tNearPlanes = (vec3(nearPlanes) - rayOrigin) / rayDirection;
            vec3 tFarPlanes  = (vec3(farPlanes ) - rayOrigin) / rayDirection;

            // Validate that the ray intersects the volume
            float tMaxOfNear = max(max(tNearPlanes.x, tNearPlanes.y), tNearPlanes.z);
            if (tMaxOfNear < 0.0) {
                // Wrong way
                return false;
            }
            float tMinOfFar  = min(min(tFarPlanes.x,  tFarPlanes.y),  tFarPlanes.z);
            if (tMaxOfNear < tMinOfFar) {
                // Missed
                return false;
            }

            // TODO: Move rayIdx to edge of volume here.
        }

        bool hit = false;
        uint i = 0u;

        do {
            vec3 tNextPlanes = (rayIdx + ivec3(positiveDir) - rayOrigin) / rayDirection;
            tNextPlanes = mix(tNextPlanes, vec3(1.0 / 0.0), equal(rayDirection, vec3(0.0)));

            float tHitPlane = min(min(tNextPlanes.x, tNextPlanes.y), tNextPlanes.z);
            bvec3 minComp = equal(tNextPlanes, vec3(tHitPlane));
            rayIdx += ivec3(mix(ivec3(0), stepv, minComp));
            hitNormal = mix(vec3(0.0), vec3(-stepv), minComp);

            if (OutsideVoxelmap(rayIdx)) { break; }

            ivec2 P = CalculateVoxelmapIdx(rayIdx);
            //hit = texelFetch(shadowtex0, P, 0).x <= (0.5 / OCTREE_LEVELS);
            float texel = unpackUnorm4x8(texelFetch(shadowcolor1, P, 0).x).w;
            hit = int(255.0 * texel + 0.5) <= (255 - OCTREE_LEVELS);
            if (hit) {
                hitPosition = rayOrigin + tHitPlane * rayDirection;
                hitIdx = rayIdx;
            }
        } while (++i < 4096u && !hit);

        return hit;
    }

    bool TraverseVoxels(vec3 rayOrigin, ivec3 rayIdx, vec3 rayDirection, out vec3 hitPosition, out ivec3 hitIdx, out vec3 hitNormal) {
        /*
        return TraverseVoxelsNooctree(
            rayOrigin, rayIdx, rayDirection,
            hitPosition, hitIdx, hitNormal
        );
        //*/

        // Validate rayDirection
        if (any(isnan(rayDirection)) || (rayDirection == vec3(0.0))) { return false; }

        // Fix inf directions
        // These probably aren't intended, but if they are they should be handled properly.
        bvec3 infComponent = isinf(rayDirection);
        if (any(infComponent)) {
            rayDirection = mix(vec3(0.0), SignExtract(rayDirection), infComponent);
        }

        vec3  pos = rayOrigin;
        ivec3 idx = rayIdx;

        ivec3 deltaSign = ivec3(sign(rayDirection));
        ivec3 positiveDir = max(deltaSign, 0);
        ivec3 negativeDir = min(deltaSign, 0);

        if (OutsideVoxelmap(idx)) {
            uvec3 plane    = minBounds.xyx - negativeDir * maxBounds.xyx;
            uvec3 farPlane = minBounds.xyx + positiveDir * maxBounds.xyx;
            vec3 planeT    = (   plane - pos) / rayDirection;
            vec3 farPlaneT = (farPlane - pos) / rayDirection;

            float maxT = max(max(planeT.x, planeT.y), planeT.z);
            if (maxT < 0.0) { return false; }
            float minFarT = min(min(farPlaneT.x, farPlaneT.y), farPlaneT.z);
            if (minFarT < maxT) { return false; }

            bvec3 isMaxComp = bvec3(planeT.x == maxT, planeT.y == maxT, planeT.z == maxT);

            pos += rayDirection * maxT;
            idx.x = isMaxComp.x ? int(plane.x) + negativeDir.x : idx.x + int(uintBitsToFloat(floatBitsToUint(pos.x + (negativeDir.x - idx.x)) - 1u));
            idx.y = isMaxComp.y ? int(plane.y) + negativeDir.y : idx.y + int(uintBitsToFloat(floatBitsToUint(pos.y + (negativeDir.y - idx.y)) - 1u));
            idx.z = isMaxComp.z ? int(plane.z) + negativeDir.z : idx.z + int(uintBitsToFloat(floatBitsToUint(pos.z + (negativeDir.z - idx.z)) - 1u));

            ivec2 P = CalculateVoxelmapIdx(idx);
            //bool hit = texelFetch(shadowtex0, P, 0).r <= (0.5 / OCTREE_LEVELS);
            float texel = unpackUnorm4x8(texelFetch(shadowcolor1, P, 0).x).w;
            bool hit = int(255.0 * texel + 0.5) <= (255 - OCTREE_LEVELS + 0);
            if (hit) {
                hitPosition = pos;
                hitIdx = idx;

                hitNormal = mix(vec3(0.0), -deltaSign, isMaxComp);

                return true;
            }
        }

        const int maxLevel = OCTREE_LEVELS - 1;
        int level = 0;

        vec3 tmp = rayOrigin + (negativeDir - rayIdx);
        float minT;
        uint minComp;
        bvec3 isMinComp = bvec3(false);
        bool hit = false;
        uint i = 0u;
        do {
            ivec3 plane = ((idx >> level) + positiveDir) << level;
            //ivec3 plane = (idx & (0xffffffff << level)) + (positiveDir << level);
            #ifdef RELIABLE_VOXEL_TRAVERSAL
                vec3 planeOffsets = plane - pos;
            #else
                vec3 planeOffsets = plane - rayOrigin;
            #endif
            vec3 planeT = abs(planeOffsets / rayDirection); // abs to avoid issues with -inf after division

            #ifdef RELIABLE_VOXEL_TRAVERSAL
                // Rays seem to never get stuck
                // The uint subtract by 1 as well as only ever having 1 component as minimum also ensure it doens't skip corners

                //* first reliable method i've come up with, also the fastest as far as i can tell
                minT = min(min(planeT.x, planeT.y), planeT.z);
                isMinComp.x = planeT.x == minT;
                isMinComp.y = !isMinComp.x && planeT.y == minT;
                isMinComp.z = !(isMinComp.x || isMinComp.y);

                pos += rayDirection * minT;
                idx.x = isMinComp.x ? plane.x + negativeDir.x : idx.x + int(uintBitsToFloat(floatBitsToUint(pos.x + (negativeDir.x - idx.x)) - 1u));
                idx.y = isMinComp.y ? plane.y + negativeDir.y : idx.y + int(uintBitsToFloat(floatBitsToUint(pos.y + (negativeDir.y - idx.y)) - 1u));
                idx.z = isMinComp.z ? plane.z + negativeDir.z : idx.z + int(uintBitsToFloat(floatBitsToUint(pos.z + (negativeDir.z - idx.z)) - 1u));
                //*/

                /* theorethically should be faster, but for some reason determining index of min component is slow?
                // the actual indexing doesn't seem to be the issue so i'm not sure
                minComp = planeT[1] < planeT[0] ? 1 : 0;
                minComp = planeT[2] < planeT[minComp] ? 2 : minComp;

                pos += rayDirection * planeT[minComp];
                idx += ivec3(uintBitsToFloat(floatBitsToUint(pos + (negativeDir - idx)) - 1u));
                idx[minComp] = plane[minComp] + negativeDir[minComp];
                //*/

                /*
                minComp = planeT[1] < planeT[0] ? 1 : 0;
                minComp = planeT[2] < planeT[minComp] ? 2 : minComp;

                pos + rayDirection * planeT[minComp];
                pos + rayDirection * abs((plane - pos) / rayDirection)[minComp];
                pos + rayDirection * abs((plane[minComp] - pos[minComp]) / rayDirection[minComp]);
                pos + rayDirection * abs(plane[minComp] - pos[minComp]) / abs(rayDirection[minComp]);

                pos += rayDirection * planeT[minComp];
                idx += ivec3(uintBitsToFloat(floatBitsToUint(pos + (negativeDir - idx)) - 1u));
                idx[minComp] = plane[minComp] + negativeDir[minComp];
                //*/
            #else
                // Slightly faster, but rays can get stuck and/or skip corners
                minT = min(min(planeT.x, planeT.y), planeT.z);
                isMinComp = bvec3(planeT.x == minT, planeT.y == minT, planeT.z == minT);

                idx.x = isMinComp.x ? plane.x + negativeDir.x : rayIdx.x + int(rayDirection.x * minT + tmp.x);
                idx.y = isMinComp.y ? plane.y + negativeDir.y : rayIdx.y + int(rayDirection.y * minT + tmp.y);
                idx.z = isMinComp.z ? plane.z + negativeDir.z : rayIdx.z + int(rayDirection.z * minT + tmp.z);
            #endif

            if (OutsideVoxelmap(idx)) { break; }

            level = min(level + 1, maxLevel);

            ivec2 P = CalculateVoxelmapIdx((idx >> level) << level);
            //hit = texelFetch(shadowtex0, P, 0).r <= ((level + 0.5) / OCTREE_LEVELS);
            float texel = unpackUnorm4x8(texelFetch(shadowcolor1, P, 0).x).w;
            hit = int(255.0 * texel + 0.5) <= (255 - OCTREE_LEVELS + level);
            while (hit && level > 0) {
                --level;
                P = CalculateVoxelmapIdx((idx >> level) << level);
                //hit = texelFetch(shadowtex0, P, 0).r <= ((level + 0.5) / OCTREE_LEVELS);
                texel = unpackUnorm4x8(texelFetch(shadowcolor1, P, 0).x).w;
                hit = int(255.0 * texel + 0.5) <= (255 - OCTREE_LEVELS + level);
            }
        } while (!hit && ++i < 4096u);

        traversalSteps += i;

        #ifdef RELIABLE_VOXEL_TRAVERSAL
        hitPosition = pos;
        #else
        hitPosition = rayOrigin + rayDirection * minT;
        #endif
        hitIdx = idx;

        //*
        if (isMinComp.x) { hitNormal = vec3(-deltaSign.x, 0.0, 0.0); }
        if (isMinComp.y) { hitNormal = vec3(0.0, -deltaSign.y, 0.0); }
        if (isMinComp.z) { hitNormal = vec3(0.0, 0.0, -deltaSign.z); }
        //*/
        /*
        hitNormal = vec3(0.0);
        hitNormal[minComp] = -deltaSign[minComp];
        //*/

        return hit;
    }
    bool RaytraceVoxel(vec3 position, ivec3 index, vec3 direction, out vec3 hitPosition, out ivec3 hitIndex, out float hitDistance, out VoxelData data) {
        //https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.42.3443&rep=rep1&type=pdf
        if(OutsideVoxelmap(index)) {
            return false;
        }

        vec4 voxel0 = unpackUnorm4x8(texelFetch(shadowcolor1, CalculateVoxelmapIdx(index), 0).r);
        vec4 voxel1 = texelFetch(shadowcolor0, CalculateVoxelmapIdx(index), 0);

        bool hit = false;
        int id = int(voxel0.a * 255.0 + 0.5);
        if(id <= (255 - OCTREE_LEVELS)) {
            if(id == 157 || id == 158 || id == 159 || id == 161) {
                if(IntersectAABoxBackFace(position - index, 1.0 / direction, vec3(0.0), vec3(1.0), voxel0, voxel1, hitDistance, data, true)) {
                    hitPosition = position + hitDistance * direction;
                    hitIndex    = index;

                    data.sameMaterial = false;

                    return true;
                }
            } else if(id == 166) {
                if(IntersectSlime(position - index, 1.0 / direction, voxel0, voxel1, hitDistance, data)) {
                    hitPosition = position + hitDistance * direction;
                    hitIndex    = index;

                    data.sameMaterial = false;

                    return true;
                }
            } else if(id == 164 || id == 165 || id == 172) {
                if(IntersectCrossBlock(position - index, direction, voxel0, voxel1, hitDistance, data)) {
                    hitPosition = position + hitDistance * direction;
                    hitIndex    = index;

                    data.sameMaterial = false;

                    return true;
                }
            } else if(id >= 2 && id <= 96) {
                if(IntersectVoxel(position - index, direction, id, voxel0, voxel1, hitDistance, data)) {
                    hitPosition = position + hitDistance * direction;
                    hitIndex    = index;

                    data.sameMaterial = false;

                    return true;
                }
            }
        }

        data.sameMaterial = false;

        hitPosition = position;

        int previousID = 0;
        int currentID  = int(voxel0.a * 255.0 + 0.5);
        for(int i = 0; i < RAYTRACE_MAX_T; ++i) {
            hit = TraverseVoxels(position, index, direction, hitPosition, hitIndex, data.flatNormal);

            if(!hit) {
                break;
            }

            hitDistance = distance(position, hitPosition);
            index = hitIndex;

            voxel0 = unpackUnorm4x8(texelFetch(shadowcolor1, CalculateVoxelmapIdx(index), 0).r);
            voxel1 = texelFetch(shadowcolor0, CalculateVoxelmapIdx(index), 0);

            previousID = currentID;
            currentID  = int(voxel0.a * 255.0 + 0.5);

            if(hit) {
                if(currentID == 157 || currentID == 158 || currentID == 159 || currentID == 161) {
                    hit = IntersectAABoxBackFace(position - hitIndex, 1.0 / direction, vec3(0.0), vec3(1.0), voxel0, voxel1, hitDistance, data, true);

                    if(hit) {
                        hitPosition = position + hitDistance * direction;

                        data.sameMaterial = previousID == currentID;

                        break;
                    }
                } else if(currentID == 166) {
                    hit = IntersectSlime(position - hitIndex, 1.0 / direction, voxel0, voxel1, hitDistance, data);

                    if(hit) {
                        hitPosition = position + hitDistance * direction;

                        data.sameMaterial = previousID == currentID;

                        break;
                    }
                } else if(currentID == 164 || currentID == 165 || currentID == 172) {
                    bool crossBlock = IntersectCrossBlock(position - hitIndex, direction, voxel0, voxel1, hitDistance, data);
                    hit = crossBlock && data.opacity > 0.102;

                    if(hit) {
                        hitPosition = position + hitDistance * direction;

                        data.sameMaterial = false;

                        break;
                    }
                } else if(currentID >= 2 && currentID <= 96) {
                    hit = IntersectVoxel(position - hitIndex, direction, currentID, voxel0, voxel1, hitDistance, data);

                    if(hit) {
                        hitPosition = position + hitDistance * direction;

                        data.sameMaterial = false;

                        break;
                    }
                } else {
                    data.id = currentID;

                    if (abs(data.flatNormal.y) > abs(data.flatNormal.x) && abs(data.flatNormal.y) > abs(data.flatNormal.z)) {
                        data.tbnMatrix = mat3(
                            vec3(1.0, 0.0, 0.0),
                            vec3(0.0, 0.0, sign(data.flatNormal.y)),
                            data.flatNormal
                        );
                    } else {
                        data.tbnMatrix = mat3(
                            vec3(sign(data.flatNormal.z), 0.0, -sign(data.flatNormal.x)),
                            vec3(0.0, -1.0, 0.0),
                            data.flatNormal
                        );
                    }

                    vec2 tileIdx = 256.0 * 15.0 * UnpackUnorm2x4(voxel1.y) + 255.0 * voxel1.zw;
                    ivec2 tileSize = ivec2(exp2(voxel1.x * 15.0));
                    ivec2 tileOffset = ivec2(tileIdx * tileSize);

                    mat2 tileInfo = mat2(
                        exp2(voxel1.x * 15.0) / textureSize(colortex6, 0),
                        tileIdx * (exp2(voxel1.x * 15.0) / textureSize(colortex6, 0))
                    );

                    data.uv  = vec2(dot(hitPosition, data.tbnMatrix[0]), dot(hitPosition, data.tbnMatrix[1]));
                    data.uv  = fract(data.uv);
                    data.uv += tileIdx;
                    data.uv *= tileInfo[0];

                    data.color = SrgbToLinear(textureLod(colortex6, data.uv, 0).rgb * voxel0.rgb);

                    vec2 rawNormal = textureLod(colortex7, data.uv, 0).xy * 2.0 - 1.0;
                    #ifndef MC_NORMAL_MAP
                        data.normalMap = normalize(data.tbnMatrix * vec3(0, 0, 1));
                        data.tangentNormal = normalize(data.normalMap * data.tbnMatrix);
                    #else
                        data.normalMap = normalize(data.tbnMatrix * vec3(rawNormal.xy, sqrt(saturate(1.0 - dot(rawNormal.xy, rawNormal.xy)))));
                        data.tangentNormal = normalize(data.normalMap * data.tbnMatrix);
                    #endif

                    data.specular = textureLod(colortex8, data.uv, 0);

                    data.ambientOcclusion = textureLod(colortex7, data.uv, 0).b;

                    data.opacity = textureLod(colortex6, data.uv, 0).a;

                    data.thickness = 0.0;

                    bool alphaTest = data.opacity > max(RandNextF(), 26.5 / 255.0);
                    if(data.id == 17) {
                        alphaTest = true;
                    }

                    data.sameMaterial = false;//previousID == currentID;

                    hit = alphaTest;

                    if(hit) {
                        break;
                    }
                }
            }
        }

        return hit;
    }
#endif