#version 450

#include "/settings.glsl"

layout (triangles) in;
layout (points, max_vertices = 1) out;

uniform sampler2D tex;

uniform vec3 cameraPosition;

uniform mat4 shadowModelViewInverse;
uniform mat4 shadowProjection;
uniform mat4 modelViewMatrix;
uniform mat4 projectionMatrix;

uniform mat3 normalMatrix;

uniform vec4 spriteBounds;

uniform ivec2 atlasSize;

in float blockID[];
in vec2 uv[];
in vec2 midCoord[];
in vec3 midBlock[];
flat in vec3 vertexNormal[];
in vec3 color[];
flat in vec4 vertexTangent[];

flat out vec4[2] voxelData;

#include "/lib/utility/universal.glsl"
#include "/lib/voxel/voxelization.glsl"

vec3 DetermineVoxelLocation() {
    vec3 tmp = gl_in[0].gl_Position.xyz + midBlock[0] / 64.0;

    tmp = mat3(modelViewMatrix) * tmp + modelViewMatrix[3].xyz;
    tmp = mat3(shadowModelViewInverse) * tmp + shadowModelViewInverse[3].xyz;
    tmp = VoxelSpaceFromSceneSpace(tmp);

    return floor(tmp);
}

vec3 FindQuadCenter() {
    vec3 m;
    m.xy = (midCoord[0] - uv[2]) * inverse(mat2(
        uv[0].x - uv[2].x, uv[1].x - uv[2].x,
        uv[0].y - uv[2].y, uv[1].y - uv[2].y
    ));
    m.z = 1.0 - m.x - m.y;

    return mat3(gl_in[0].gl_Position.xyz, gl_in[1].gl_Position.xyz, gl_in[2].gl_Position.xyz) * m;
}

void main() {
    if(blockID[0] == 0) {
        return;
    }

    vec3 normal = mat3(shadowModelViewInverse) * normalize(normalMatrix * vertexNormal[0]);

    if (((blockID[0] >= 51.0 && blockID[0] <= 57.0) || (blockID[0] >= 80.0 && blockID[0] <= 88.0) || (blockID[0] >= 230 && blockID[0] <= 245.0)) && abs(normal.y) > 0.9) {
        return;
    }

    vec3 pos = FindQuadCenter();
         pos = transMAD(modelViewMatrix, pos);
         pos = transMAD(shadowModelViewInverse, pos);

    ivec3 voxelIndex = ivec3(DetermineVoxelLocation());
    vec4 p2d = vec4(((CalculateVoxelmapIdx(voxelIndex) + 0.5) / float(shadowMapQuality)) * 2.0 - 1.0, vertexNormal[0].y * -0.25 + 0.5, 1.0);

    if(!InsideVoxelmap(voxelIndex)) {
        return;
    }

    vec3 bitangent = cross(vertexTangent[0].xyz, vertexNormal[0]) * (vertexTangent[0].w < 0.0 ? -1.0 : 1.0);
    vec3 vertexPositionU = vertexTangent[0].xyz * mat3(gl_in[0].gl_Position.xyz, gl_in[1].gl_Position.xyz, gl_in[2].gl_Position.xyz);
    vec3 vertexPositionV = bitangent            * mat3(gl_in[0].gl_Position.xyz, gl_in[1].gl_Position.xyz, gl_in[2].gl_Position.xyz);
    vec2 faceSize  = vec2(maxof(vertexPositionU) - minof(vertexPositionU), maxof(vertexPositionV) - minof(vertexPositionV));
         faceSize /= vec2(length(vertexTangent[0].xyz), length(bitangent));

    #if defined MC_ANISOTROPIC_FILTERING
        ivec2 tileRes = findMSB(textureSize(tex, 0));
        vec2 tileSize = spriteBounds.zw - spriteBounds.xy;
        ivec2 tileIdx = ivec2(round(spriteBounds.xy / tileSize));

        ivec2 tileResolution = ivec2(round(tileSize * atlasSize));
    #else
        vec2 minCoord = min(min(uv[0], uv[1]), uv[2]);
        vec2 maxCoord = max(max(uv[0], uv[1]), uv[2]);
        vec2 tileSize = abs(maxCoord - minCoord) / faceSize;
        ivec2 tileResolution = ivec2(round(tileSize * atlasSize));
        ivec2 tileIdx = ivec2(midCoord[0] / tileSize);

        if(blockID[0] == 164 || blockID[0] == 172 || blockID[0] == 173) {
            vec2 tileSize = 2.0 * abs(uv[0] - midCoord[0]);

            tileResolution = ivec2(round(tileSize * atlasSize));
            tileIdx = ivec2(midCoord[0] / tileSize);
        }
    #endif

    voxelData[0] = vec4(color[0], min(blockID[0], 230.0) / 255.0);
    voxelData[1] = vec4(log2(maxof(tileResolution)) / 15.0, PackUnorm2x4(floor(tileIdx / 256.0) / 15.0), fract(tileIdx / 256.0) * 256.0 / 255.0);

    float normalBias = 0.25 - 0.125 * normal.y;

    ivec2 P = CalculateVoxelmapIdx(voxelIndex);

    gl_Position = vec4(((P + 0.5) / shadowMapQuality) * 2.0 - 1.0, normalBias, 1.0);
    gl_PointSize = 1.0;

    EmitVertex();
    EndPrimitive();
}